#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.bds.big_data_service import ADSHiveConnection
from ads.common.decorator.runtime_dependency import OptionalDependency
from typing import Dict, Optional, Union, Iterator
from pandas import DataFrame


class ConnectionFactory:
    connectionprovider = {"hive": ADSHiveConnection}

    try:
        from ads.oracledb.oracle_db import OracleRDBMSConnection

        connectionprovider["oracle"] = OracleRDBMSConnection
    except:
        pass

    try:
        from ads.mysqldb.mysql_db import MySQLRDBMSConnection

        connectionprovider["mysql"] = MySQLRDBMSConnection
    except:
        pass

    @classmethod
    def get(cls, engine="oracle"):
        Connection = cls.connectionprovider.get(engine, None)

        if not Connection:
            if engine == "mysql":
                print("Requires mysql-connector-python package to use mysql engine")
            elif engine == "oracle":
                print(
                    f"The `oracledb` or `cx_Oracle` module was not found. Please run "
                    f"`pip install {OptionalDependency.DATA}`."
                )
            raise Exception(
                f"Engine {engine} does not have either required dependency or is not supported."
            )
        return Connection


class DBAccessMixin:
    @staticmethod
    def read_sql(
        sql: str,
        connection_parameters: dict,
        bind_variables: Dict = {},
        chunksize: Optional[int] = None,
        engine="oracle",
    ) -> Union["DataFrame", Iterator["DataFrame"]]:
        """Read SQL query from oracle database into a DataFrame.

        Parameters
        ----------
        sql: str
            SQL query to be executed.
        connection_parameters: dict
            A dictionary of connection_parameters - {"user_name":"", "password":"", "service_name":"", "wallet_location":""}
        bind_variables: Optional[Dict]
            Key value of pair of bind variables and corresponding values
        chunksize: Optional[int], default None
            If specified, return an iterator where `chunksize` is the number of rows to include in each chunk.
        engine: {'oracle', 'mysql', 'hive'}, default 'oracle'
            Select the database type - MySQL/Oracle/Hive to store the data

        Returns
        -------
            DataFrame or Iterator[DataFrame]
                DataFrame or Iterator[DataFrame].

        Examples
        --------
        >>> connection_parameters = {
                "user_name": "<username>",
                "password": "<password>",
                "service_name": "{service_name}_{high|med|low}",
                "wallet_location": "/full/path/to/my_wallet.zip",
            }
        >>> import pandas as pd
        >>> import ads
        >>> df = pd.DataFrame.ads.read_sql("SELECT * from Employee", connection_parameters=connection_parameters)
        >>> df_with_bind = pd.DataFrame.ads.read_sql("SELECT * from EMPLOYEE WHERE EMPLOYEE_ID = :ID", bind_variables={"ID":"121212", connection_parameters=connection_parameters)


        """
        Connection = ConnectionFactory.get(engine)

        return Connection(**connection_parameters).query(
            sql, bind_variables=bind_variables, chunksize=chunksize
        )

    def to_sql(
        self,
        table_name: str,
        connection_parameters: dict,
        if_exists: str = "fail",
        batch_size=100000,
        engine="oracle",
        encoding="utf-8",
    ):
        """To save the dataframe df to database.

        Parameters
        ----------
        table_name: str
            Name of SQL table.
        connection_parameters: dict
            A dictionary of connection_parameters - {"user_name":"", "password":"", "service_name":"", "wallet_location":""}
        if_exists: : {'fail', 'replace', 'append'}, default 'fail'
            How to behave if the table already exists.
            * fail: Raise a ValueError. If table exists, do nothing
            * replace: Drop the table before inserting new values. If table exists, drop it, recreate it, and insert data.
            * append: Insert new values to the existing table. If table exists, insert data. Create if does not exist.
        batch_size: int, default 100000
            Inserting in batches improves insertion performance. Choose this value based on available memore and network bandwidth.
        engine: {'oracle', 'mysql'}, default 'oracle'
            Select the database type - MySQL or Oracle to store the data
        encoding: str, default is "utf-8"
            Encoding provided will be used for ecoding all columns, when inserting into table


        Returns
        -------
            None
                Nothing.
        Examples
        --------
        >>> connection_parameters = {
                "user_name": "<username>",
                "password": "<password>",
                "service_name": "{service_name}_{high|med|low}",
                "wallet_location": "/full/path/to/my_wallet.zip",
            }
        >>> import pandas as pd
        >>> import ads
        >>> df2 = pd.read_csv("my/data/csv")
        >>> df2.ads.to_sql("MY_DATA_CSV", connection_parameters=connection_parameters)
        """
        if if_exists not in ["fail", "replace", "append"]:
            raise ValueError(
                f"Unknown option `if_exists`={if_exists}. Valid options are 'fail', 'replace', 'append'"
            )

        Connection = ConnectionFactory.get(engine)
        return Connection(**connection_parameters).insert(
            table_name, self._obj, if_exists, batch_size, encoding
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import logging
import os
import sys
import textwrap
from typing import List, Tuple, Union

import git
import numpy as np
import pandas as pd
from ads.common import logger, utils
from ads.common.data import ADSData
from ads.common.model_metadata import (
    MetadataCustomCategory,
    MetadataCustomKeys,
    MetadataTaxonomyKeys,
    ModelCustomMetadataItem,
    ModelProvenanceMetadata,
    UseCaseType,
)
from ads.common.utils import DATA_SCHEMA_MAX_COL_NUM, get_files
from ads.feature_engineering.schema import DataSizeTooWide, Schema, SchemaSizeTooLarge

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("ADS")

METADATA_SIZE_LIMIT = 32000
METADATA_VALUE_LENGTH_LIMIT = 255
METADATA_DESCRIPTION_LENGTH_LIMIT = 255

MODEL_ARTIFACT_VERSION = "3.0"
INPUT_SCHEMA_FILE_NAME = "input_schema.json"
OUTPUT_SCHEMA_FILE_NAME = "output_schema.json"
ADS_PACKAGE_NAME = "ADS"


class MetadataMixin:
    """MetadataMixin class which populates the custom metadata, taxonomy metadata,
    input/output schema and provenance metadata.
    """

    def _populate_metadata_taxonomy(
        self, model: callable = None, use_case_type: str = None
    ):
        """Populates the taxonomy metadata.

        Parameters
        ----------
        model: (callable, optional). Defaults to None.
            Any model.
        use_case_type: (str, optional). Defaults to None.
            The use case type of the model.

        Raises
        ------
        ValueError: if `UseCaseType` is not from UseCaseType.values().

        Returns
        -------
        None
            Nothing.
        """
        if use_case_type and use_case_type not in UseCaseType:
            raise ValueError(
                f"Invalid value of `UseCaseType`. Choose from {UseCaseType.values()}."
            )

        self.metadata_taxonomy[MetadataTaxonomyKeys.USE_CASE_TYPE].value = use_case_type
        if model is not None:
            mapping = {
                MetadataTaxonomyKeys.FRAMEWORK: self.framework,
                MetadataTaxonomyKeys.FRAMEWORK_VERSION: self.version,
                MetadataTaxonomyKeys.ALGORITHM: str(self.algorithm),
                MetadataTaxonomyKeys.HYPERPARAMETERS: self.hyperparameter,
            }

            if mapping is not None:
                self.metadata_taxonomy._populate_from_map(mapping)
            if (
                self.metadata_taxonomy[MetadataTaxonomyKeys.HYPERPARAMETERS].size()
                > METADATA_SIZE_LIMIT
            ):
                logger.warn(
                    f"The model hyperparameters are larger than `{METADATA_SIZE_LIMIT}` "
                    "bytes and cannot be stored as model catalog metadata. It will be saved to "
                    f"{self.artifact_dir}/hyperparameters.json and removed from the metadata."
                )

                self.metadata_taxonomy[
                    MetadataTaxonomyKeys.HYPERPARAMETERS
                ].to_json_file(self.artifact_dir)
                self.metadata_taxonomy[MetadataTaxonomyKeys.HYPERPARAMETERS].update(
                    value=None
                )

    def _populate_metadata_custom(self):
        """Populates the custom metadata.

        Returns
        -------
        None
            Nothing.
        """
        model_metadata_items = []

        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CONDA_ENVIRONMENT,
                value=self.runtime_info.model_provenance.training_conda_env.training_env_path,
                description="The conda environment where the model was trained.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )

        try:
            env_type = (
                self.runtime_info.model_provenance.training_conda_env.training_env_type
            )
        except:
            env_type = None
        try:
            slug_name = (
                self.runtime_info.model_provenance.training_conda_env.training_env_slug
            )
        except:
            slug_name = None
        try:
            env_path = (
                self.runtime_info.model_provenance.training_conda_env.training_env_path
            )
        except:
            env_path = None

        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.ENVIRONMENT_TYPE,
                value=env_type,
                description="The conda environment type, can be published or datascience.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.SLUG_NAME,
                value=slug_name,
                description="The slug name of the training conda environment.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CONDA_ENVIRONMENT_PATH,
                value=env_path,
                description="The URI of the training conda environment.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )
        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.MODEL_ARTIFACTS,
                value=textwrap.shorten(
                    ", ".join(get_files(self.artifact_dir)),
                    255,
                    placeholder="...",
                ),
                description="The list of files located in artifacts folder.",
                category=MetadataCustomCategory.TRAINING_ENV,
            )
        )

        if self.model_file_name:
            model_metadata_items.append(
                ModelCustomMetadataItem(
                    key=MetadataCustomKeys.MODEL_SERIALIZATION_FORMAT,
                    value=self.model_file_name.split(".")[-1],
                    description="The model serialization format.",
                    category=MetadataCustomCategory.TRAINING_PROFILE,
                )
            )
            model_metadata_items.append(
                ModelCustomMetadataItem(
                    key=MetadataCustomKeys.MODEL_FILE_NAME,
                    value=self.model_file_name,
                    description="The model file name.",
                    category=MetadataCustomCategory.OTHER,
                )
            )
        else:
            logger.warning(
                "Unable to extract a model serialization format. "
                "The `model_file_name` is not provided."
            )

        model_metadata_items.append(
            ModelCustomMetadataItem(
                key=MetadataCustomKeys.CLIENT_LIBRARY,
                value=ADS_PACKAGE_NAME,
                description="",
                category=MetadataCustomCategory.OTHER,
            )
        )
        self.metadata_custom._add_many(model_metadata_items, replace=True)

    def _prepare_data_for_schema(
        self,
        X_sample: Union[List, Tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        y_sample: Union[List, Tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
    ):
        """
        Any Framework-specific work before generic schema generation.
        """
        return X_sample, y_sample

    def _populate_provenance_metadata(
        self,
        training_script_path: str = "",
        training_id: str = None,
        ignore_pending_changes: bool = False,
    ):
        """Populates the custom metadata.

        Parameters
        ----------
        training_script_path: (str, optional). Defaults to "".
            Training script path.
        training_id: (str, optional). Defaults to None.
            The training model OCID.
        ignore_pending_changes: (bool, optional). Defaults to False.
            Ignore the pending changes in git.

        Returns
        -------
        None
            Nothing.
        """
        try:
            self.metadata_provenance = (
                ModelProvenanceMetadata.fetch_training_code_details(
                    training_script_path=training_script_path,
                    training_id=training_id,
                    artifact_dir=self.artifact_dir,
                )
            )
        except git.InvalidGitRepositoryError:
            self.metadata_provenance = ModelProvenanceMetadata()

        if training_script_path is not None:
            # training_script_path could be a directory or a path.
            # but if the path has too many files, it might take a long time.
            self.metadata_provenance.assert_path_not_dirty(
                training_script_path, ignore_pending_changes
            )
        self.metadata_provenance.assert_path_not_dirty(
            self.artifact_dir, ignore_pending_changes
        )

    def populate_metadata(
        self,
        use_case_type: str = None,
        data_sample: ADSData = None,
        X_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        y_sample: Union[list, tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        training_script_path: str = None,
        training_id: str = None,
        ignore_pending_changes: bool = True,
        max_col_num: int = DATA_SCHEMA_MAX_COL_NUM,
    ):
        """Populates input schema and output schema.
        If the schema exceeds the limit of 32kb, save as json files to the artifact directory.

        Parameters
        ----------
        use_case_type: (str, optional). Defaults to None.
            The use case type of the model.
        data_sample: (ADSData, optional). Defaults to None.
            A sample of the data that will be used to generate intput_schema and output_schema.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]. Defaults to None.
            A sample of input data that will be used to generate input schema.
        y_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]. Defaults to None.
            A sample of output data that will be used to generate output schema.
        training_script_path: str. Defaults to None.
            Training script path.
        training_id: (str, optional). Defaults to None.
            The training model OCID.
        ignore_pending_changes: bool. Defaults to False.
            Ignore the pending changes in git.
        max_col_num: (int, optional). Defaults to utils.DATA_SCHEMA_MAX_COL_NUM.
            The maximum number of columns allowed in auto generated schema.

        Returns
        -------
        None
            Nothing.
        """
        if (
            self.estimator is None
            and self.metadata_taxonomy[MetadataTaxonomyKeys.ALGORITHM].value is None
        ):
            logger.info(
                "To auto-extract taxonomy metadata the model must be provided. Supported models: automl, keras, lightgbm, pytorch, sklearn, tensorflow, pyspark, and xgboost."
            )
        if use_case_type is None:
            use_case_type = self.metadata_taxonomy[
                MetadataTaxonomyKeys.USE_CASE_TYPE
            ].value
        self._populate_metadata_taxonomy(
            model=self.estimator, use_case_type=use_case_type
        )
        self._populate_metadata_custom()
        self.populate_schema(
            data_sample=data_sample,
            X_sample=X_sample,
            y_sample=y_sample,
            max_col_num=max_col_num,
        )
        self._populate_provenance_metadata(
            training_script_path=training_script_path,
            ignore_pending_changes=ignore_pending_changes,
            training_id=training_id,
        )
        self._summary_status.update_action(
            detail="Populated metadata(Custom, Taxonomy and Provenance)", action=""
        )
        self._summary_status.update_status(
            detail="Populated metadata(Custom, Taxonomy and Provenance)",
            status="Done",
        )

    def populate_schema(
        self,
        data_sample: ADSData = None,
        X_sample: Union[List, Tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        y_sample: Union[List, Tuple, pd.DataFrame, pd.Series, np.ndarray] = None,
        max_col_num: int = DATA_SCHEMA_MAX_COL_NUM,
    ):
        """Populate input and output schemas.
        If the schema exceeds the limit of 32kb, save as json files to the artifact dir.

        Parameters
        ----------
        data_sample: ADSData
            A sample of the data that will be used to generate input_schema and output_schema.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
            A sample of input data that will be used to generate the input schema.
        y_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
            A sample of output data that will be used to generate the output schema.
        max_col_num: (int, optional). Defaults to utils.DATA_SCHEMA_MAX_COL_NUM.
            The maximum number of columns allowed in auto generated schema.
        """
        if data_sample is not None:
            assert isinstance(
                data_sample, ADSData
            ), "`data_sample` expects data of ADSData type. \
            Pass in to `X_sample` and `y_sample` for other data types."
            X_sample = data_sample.X
            y_sample = data_sample.y

        X_sample, y_sample = self._prepare_data_for_schema(X_sample, y_sample)
        self.schema_input = self._populate_schema(
            X_sample,
            schema_file_name=INPUT_SCHEMA_FILE_NAME,
            max_col_num=max_col_num,
        )
        self.schema_output = self._populate_schema(
            y_sample,
            schema_file_name=OUTPUT_SCHEMA_FILE_NAME,
            max_col_num=max_col_num,
        )

    def _populate_schema(
        self,
        data: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame],
        schema_file_name: str,
        max_col_num: int,
    ):
        """Populates schema and if the schema exceeds the limit of 32kb, save as a json file to the artifact_dir.

        Parameters
        ----------
        data: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
            A sample of input data that will be used to generate input schema.
        schema_file_name: str
            Filename of the schema file.
        max_col_num : int
            The maximum number of columns allowed in auto generated schema.

        Returns
        -------
        Schema
            The schema.
        """
        result = None

        try:
            if data is not None:
                data = utils.to_dataframe(data)
                schema = data.ads.model_schema(max_col_num=max_col_num)
                schema.to_json_file(os.path.join(self.artifact_dir, schema_file_name))
                if self._validate_schema_size(schema, schema_file_name):
                    result = schema
        except DataSizeTooWide:
            logger.warning(
                f"The data has too many columns and "
                f"the maximum allowable number of columns is `{max_col_num}`. "
                "The schema was not auto generated. increase allowable number of columns."
            )
        except NotImplementedError:
            logger.warning(
                f"Cannot convert the data to pandas dataframe, hence "
                "the schema was not auto generated."
            )

        return result or Schema()

    def _validate_schema_size(self, schema: Schema, schema_file_name: str):
        """Validate the schema size.

        Parameters
        ----------
        schema: (Schema)
            input/output schema.
        schema_file_name: (str)
            Filename of schema.

        Returns
        -------
        bool
            Whether the size of the schema is less than 32kb or not.
        """
        result = False
        try:
            result = schema.validate_size()
        except SchemaSizeTooLarge:
            logger.warn(
                f"The {schema_file_name.replace('.json', '')} is larger than "
                f"`{METADATA_SIZE_LIMIT}` bytes and cannot be stored as model catalog metadata."
                f"It will be saved to {self.artifact_dir}/{schema_file_name}."
            )

        return result



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import warnings

warnings.warn(
    (
        "The `ads.common.model_export_util` is deprecated in `oracle-ads 2.6.9` and will be removed in `oracle-ads 3.0`. "
        "Use framework specific Model utility class for saving and deploying model. "
        "Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html"
    ),
    DeprecationWarning,
    stacklevel=2,
)

import json
import logging
import os
import sys
from typing import Any, Dict, Union

import cloudpickle
import numpy as np
import pandas as pd
from ads.common import logger, utils
from ads.common.data import ADSData
from ads.common.decorator.runtime_dependency import (
    OptionalDependency,
    runtime_dependency,
)
from ads.common.decorator.deprecate import deprecated
from ads.common.function.fn_util import (
    generate_fn_artifacts,
    get_function_config,
    write_score,
)
from ads.common.model_artifact import ModelArtifact
from ads.model.model_metadata import UseCaseType
from ads.feature_engineering.schema import DataSizeTooWide
from pkg_resources import DistributionNotFound, get_distribution

from ads.model.transformer.onnx_transformer import (
    ONNXTransformer,
)  # Do not remove this line till ADS 3.0 wll be deployed.

pd.options.mode.chained_assignment = None

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("ADS")

OnnxConvertibleModels = [
    "sklearn",
    "automl",
    "lightgbm",
    "xgboost",
    "torch",
    "tensorflow",
    "keras",
    "mxnet",
]
NoVerifyModels = ["automl", "torch", "mxnet", "lightgbm", "xgboost"]
AlreadyWrittenModels = ["torch", "mxnet", "automl"]
TransformableData = ["automl"]
Progress_Steps_W_Fn = 6
Progress_Steps_Wo_Fn = 4


@deprecated(
    "2.6.6",
    details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
)
def prepare_generic_model(
    model_path: str,
    fn_artifact_files_included: bool = False,
    fn_name: str = "model_api",
    force_overwrite: bool = False,
    model: Any = None,
    data_sample: ADSData = None,
    use_case_type=None,
    X_sample: Union[
        list,
        tuple,
        pd.Series,
        np.ndarray,
        pd.DataFrame,
    ] = None,
    y_sample: Union[
        list,
        tuple,
        pd.Series,
        np.ndarray,
        pd.DataFrame,
    ] = None,
    **kwargs,
) -> ModelArtifact:
    """
    Generates template files to aid model deployment.
    The model could be accompanied by other artifacts all of which can be dumped at `model_path`.
    Following files are generated:
    * func.yaml
    * func.py
    * requirements.txt
    * score.py

    Parameters
    ----------
    model_path : str
        Path where the artifacts must be saved.
        The serialized model object and any other associated files/objects must
        be saved in the `model_path` directory
    fn_artifact_files_included : bool
        Default is False, if turned off, function artifacts are not generated.
    fn_name : str
        Opional parameter to specify the function name
    force_overwrite : bool
        Opional parameter to specify if the model_artifact should overwrite the existing model_path (if it exists)
    model : (Any, optional). Defaults to None.
        This is an optional model object which is only used to extract taxonomy metadata.
        Supported models: automl, keras, lightgbm, pytorch, sklearn, tensorflow, and xgboost.
        If the model is not under supported frameworks, then extracting taxonomy metadata will be skipped.
        The alternative way is using `atifact.populate_metadata(model=model, usecase_type=UseCaseType.REGRESSION)`.
    data_sample : ADSData
        A sample of the test data that will be provided to predict() API of scoring script
        Used to generate schema_input and schema_output
    use_case_type: str
        The use case type of the model
    X_sample : Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame, dask.dataframe.core.Series, dask.dataframe.core.DataFrame]
        A sample of input data that will be provided to predict() API of scoring script
        Used to generate input schema.
    y_sample : Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame, dask.dataframe.core.Series, dask.dataframe.core.DataFrame]
        A sample of output data that is expected to be returned by predict() API of scoring script,
        corresponding to X_sample
        Used to generate output schema.

    **kwargs
    ________
    data_science_env : bool, default: False
        If set to True, the datascience environment represented by the slug in the training conda environment will be used.
    inference_conda_env : str, default: None
        Conda environment to use within the model deployment service for inferencing. For example, oci://bucketname@namespace/path/to/conda/env
    ignore_deployment_error : bool, default: False
        If set to True, the prepare method will ignore all the errors that may impact model deployment.
    underlying_model : str, default: 'UNKNOWN'
        Underlying Model Type, could be "automl", "sklearn", "h2o", "lightgbm", "xgboost", "torch", "mxnet", "tensorflow", "keras", "pyod" and etc.
    model_libs : dict, default: {}
        Model required libraries where the key is the library names and the value is the library versions.
        For example, {numpy: 1.21.1}.
    progress : int, default: None
        max number of progress.
    inference_python_version: str, default:None.
        If provided will be added to the generated runtime yaml
    max_col_num: (int, optional). Defaults to utils.DATA_SCHEMA_MAX_COL_NUM.
        The maximum column size of the data that allows to auto generate schema.

    Examples
    --------
    >>> import cloudpickle
    >>> import os
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.datasets import make_classification
    >>> import ads
    >>> from ads.common.model_export_util import prepare_generic_model
    >>> import yaml
    >>> import oci
    >>>
    >>> ads.set_auth('api_key', oci_config_location=oci.config.DEFAULT_LOCATION, profile='DEFAULT')
    >>> model_artifact_location = os.path.expanduser('~/myusecase/model/')
    >>> inference_conda_env="oci://my-bucket@namespace/conda_environments/cpu/Data_Exploration_and_Manipulation_for_CPU_Python_3.7/2.0/dataexpl_p37_cpu_v2"
    >>> inference_python_version = "3.7"
    >>> if not os.path.exists(model_artifact_location):
    ...     os.makedirs(model_artifact_location)
    >>> X, y = make_classification(n_samples=100, n_features=20, n_classes=2)
    >>> lrmodel = LogisticRegression().fit(X, y)
    >>> with open(os.path.join(model_artifact_location, 'model.pkl'), "wb") as mfile:
    ...     cloudpickle.dump(lrmodel, mfile)
    >>> modelartifact = prepare_generic_model(
    ...     model_artifact_location,
    ...     model = lrmodel,
    ...     force_overwrite=True,
    ...     inference_conda_env=inference_conda_env,
    ...     ignore_deployment_error=True,
    ...     inference_python_version=inference_python_version
    ... )
    >>> modelartifact.reload() # Call reload to update the ModelArtifact object with the generated score.py
    >>> assert len(modelartifact.predict(X[:5])['prediction']) == 5 #Test the generated score.py works. This may require customization.
    >>> with open(os.path.join(model_artifact_location, "runtime.yaml")) as rf:
    ...     content = yaml.load(rf, Loader=yaml.FullLoader)
    ...     assert content['MODEL_DEPLOYMENT']['INFERENCE_CONDA_ENV']['INFERENCE_ENV_PATH'] == inference_conda_env
    ...     assert content['MODEL_DEPLOYMENT']['INFERENCE_CONDA_ENV']['INFERENCE_PYTHON_VERSION'] == inference_python_version
    >>> # Save Model to model artifact
    >>> ocimodel = modelartifact.save(
    ...     project_id="oci1......", # OCID of the project to which the model to be associated
    ...     compartment_id="oci1......", # OCID of the compartment where the model will reside
    ...     display_name="LRModel_01",
    ...     description="My Logistic Regression Model",
    ...     ignore_pending_changes=True,
    ...     timeout=100,
    ...     ignore_introspection=True,
    ... )
    >>> print(f"The OCID of the model is: {ocimodel.id}")

    Returns
    -------
    model_artifact: ads.model_artifact.model_artifact
        A generic model artifact
    """
    if "function_artifacts" in kwargs:
        if (
            fn_artifact_files_included
            or fn_artifact_files_included != kwargs["function_artifacts"]
        ):
            raise ValueError(
                "Parameter 'function_artifacts' and 'fn_artifact_files_included' cannot be used at "
                "the same time. Parameter 'function_artifacts' is deprecated and removed in future releases."
            )
        else:
            logger.warning(
                "Parameter 'function_artifacts' is deprecated and removed in future releases. Use 'fn_artifact_files_included' instead."
            )
            fn_artifact_files_included = kwargs["function_artifacts"]

    assert model_path, "Required folder path for saving artifacts"

    # If this is being called from ADSModel.prepare, grab relevant data
    underlying_model = kwargs.get("underlying_model", "UNKNOWN")
    model_libs = kwargs.get(
        "model_libs", {}
    )  # utils.extract_lib_dependencies_from_model(self.est)
    progress = kwargs.get("progress", None)
    max_col_num = kwargs.get("max_col_num", utils.DATA_SCHEMA_MAX_COL_NUM)
    artifact_type_generic = progress is not None

    from ads.common.model import ADSModel

    if isinstance(model, ADSModel) and underlying_model != "automl":
        raise ValueError(
            "Only generic model can be used to generate generic model artifact."
        )

    if use_case_type and use_case_type not in UseCaseType:
        raise ValueError(f"Invalid usecase type. Choose from {UseCaseType.values()}")

    with progress if artifact_type_generic else utils.get_progress_bar(
        Progress_Steps_W_Fn if fn_artifact_files_included else Progress_Steps_Wo_Fn
    ) as progress:
        if not artifact_type_generic:
            progress.update("Preparing Model Artifact Directory")
            if os.path.exists(model_path):
                if force_overwrite:
                    logger.warning(
                        f"As force_overwrite is set to True, all the existing files in the {model_path} will be removed"
                    )
                else:
                    raise ValueError(
                        "Directory already exists, set force_overwrite to True if you wish to overwrite."
                    )

            os.makedirs(model_path, exist_ok=True)

        progress.update("Updating requirements.txt")
        if fn_artifact_files_included:
            # fdk removed from dependency list in setup.py (fn deployments deprecated)
            # before we request versions we want to check if fdk installed by user
            # and provide support in error message, if not installed
            try:
                get_distribution("fdk")
            except Exception as e:
                if isinstance(e, DistributionNotFound):
                    error_message = (
                        "fdk library not installed in current environment, it is required "
                        "for deployment with fn. Install fdk with 'pip install fdk'."
                    )
                    logger.error(str(error_message))
                    raise
            else:
                required_fn_libs = get_function_config()["requires"]["functions"]
                [
                    model_libs.update({lib: get_distribution(lib).version})
                    for lib in required_fn_libs
                ]
                required_model_libs = get_function_config()["requires"][
                    kwargs.get("serializer", "default")
                ]
                [
                    model_libs.update({lib: get_distribution(lib).version})
                    for lib in required_model_libs
                ]
                utils.generate_requirement_file(
                    requirements=model_libs, file_path=model_path
                )

        model_artifact_args = {}
        if "inference_conda_env" in kwargs:
            model_artifact_args["inference_conda_env"] = kwargs["inference_conda_env"]
        if "inference_python_version" in kwargs:
            model_artifact_args["inference_python_version"] = kwargs[
                "inference_python_version"
            ]
        if "data_science_env" in kwargs:
            model_artifact_args["data_science_env"] = kwargs["data_science_env"]
        if "ignore_deployment_error" in kwargs:
            model_artifact_args["ignore_deployment_error"] = kwargs[
                "ignore_deployment_error"
            ]
        model_artifact = ModelArtifact(
            model_path,
            reload=False,
            create=True,
            progress=progress,
            **model_artifact_args,
        )

        model_kwargs = {"_underlying_model": underlying_model, "progress": progress}
        kwargs.update(model_kwargs)
        if fn_artifact_files_included:
            generate_fn_artifacts(
                model_path,
                fn_name,
                artifact_type_generic=artifact_type_generic,
                **kwargs,
            )
        if progress:
            progress.update("Writing score.py")
        write_score(model_path, **kwargs)
        if model is None:
            logger.warning(
                "Taxonomy metadata was not extracted. "
                "To auto-populate taxonomy metadata the model must be provided. "
                "Pass the model as a parameter to .prepare_generic_model(model=model, usecase_type=UseCaseType.REGRESSION). "
                "Alternative way is using atifact.populate_metadata(model=model, usecase_type=UseCaseType.REGRESSION)."
            )
        try:
            model_artifact.populate_metadata(model, use_case_type)
        except:
            logger.warning("Failed to populate the custom and taxonomy metadata.")

        try:
            model_artifact.populate_schema(
                data_sample,
                X_sample,
                y_sample,
                max_col_num,
            )
        except DataSizeTooWide:
            logger.warning(
                f"The data has too many columns and "
                f"the maximum allowable number of columns is `{max_col_num}`. "
                "The schema was not auto generated. increase allowable number of columns."
            )

        return model_artifact


@deprecated(
    "2.6.6",
    details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
)
def serialize_model(
    model=None, target_dir=None, X=None, y=None, model_type=None, **kwargs
):
    """
    Parameters
    ----------
    model : ads.Model
        A model to be serialized
    target_dir : str, optional
        directory to output the serialized model
    X : Union[pandas.DataFrame, pandas.Series]
        The X data
    y : Union[list, pandas.DataFrame, pandas.Series]
        Tbe Y data
    model_type : str, optional
        A string corresponding to the model type

    Returns
    -------
    model_kwargs: Dict
        A dictionary of model kwargs for the serialized model
    """
    model_kwargs = {}
    try:
        assert model_type in OnnxConvertibleModels
        assert X is not None, (
            "WARNING: In order to convert model to onnnx format, you will need to provide a data "
            "sample "
        )
        # Try to serialize the model based off of type
        onx = None
        if model_type == "sklearn":
            onx = _sklearn_to_onnx(model=model)
        elif model_type == "automl":
            _automl_to_pkl(model=model, target_dir=target_dir)
        elif model_type == "lightgbm":
            onx = _lightgbm_to_onnx(model=model, X=X, y=y)
        elif model_type == "torch":
            _torch_to_onnx(model=model, target_dir=target_dir, X=X, y=y)
        elif model_type == "xgboost":
            onx = _xgboost_to_onnx(model=model, X=X, y=y)
        elif model_type == "tensorflow":
            onx = _tf_to_onnx(model=model, X=X, y=y)
        elif model_type == "keras":
            onx = _keras_to_onnx(model=model, X=X, y=y)
        elif model_type == "mxnet":
            _mxnet_to_onnx(model=model, X=X, y=y)
        else:
            raise Exception(
                f"ADS does not have method to convert models of type: {model_type} to onnx"
            )

        # Save the model to the target file
        if model_type not in AlreadyWrittenModels:
            with open(os.path.join(target_dir, "model.onnx"), "wb") as f:
                f.write(onx.SerializeToString())

        # Lastly, set the necessary env variables for the remaining workflow
        if model_type == "automl":
            model_kwargs["model_name"] = "model.pkl"
            model_kwargs["serializer"] = "pkl"
        else:
            model_kwargs["model_name"] = "model.onnx"
            model_kwargs["serializer"] = "onnx"
    except Exception as e:
        logger.error(
            f"Failed to serialize the model as ONNX returned the error message::{e}."
        )
    return model_kwargs


@runtime_dependency(module="skl2onnx", install_from=OptionalDependency.ONNX)
@runtime_dependency(module="onnxmltools", install_from=OptionalDependency.ONNX)
@runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
def _sklearn_to_onnx(model=None, target_dir=None, X=None, y=None, **kwargs):
    from skl2onnx.common.data_types import FloatTensorType

    n_cols = model.est.n_features_in_
    initial_types = [("input", FloatTensorType([None, n_cols]))]
    return onnxmltools.convert_sklearn(
        model.est,
        name=None,
        initial_types=initial_types,
        doc_string="",
        target_opset=None,
        targeted_onnx=onnx.__version__,
        custom_conversion_functions=None,
        custom_shape_calculators=None,
    )


def _automl_to_pkl(model=None, target_dir=None, **kwargs):

    with open(os.path.join(target_dir, "model.pkl"), "wb") as outfile:
        cloudpickle.dump(model, outfile)


@runtime_dependency(module="lightgbm", install_from=OptionalDependency.BOOSTED)
@runtime_dependency(module="skl2onnx", install_from=OptionalDependency.ONNX)
@runtime_dependency(module="onnxmltools", install_from=OptionalDependency.ONNX)
@runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
def _lightgbm_to_onnx(model=None, target_dir=None, X=None, y=None, **kwargs):
    from skl2onnx.common.data_types import FloatTensorType

    in_types = [("input", FloatTensorType([None, len(X.columns)]))]

    if str(type(model.est)) == "<class 'sklearn.pipeline.Pipeline'>":
        model_est_types = [type(val[1]) for val in model.est.steps]
    else:
        model_est_types = [type(model.est)]

    if lightgbm.sklearn.LGBMClassifier in model_est_types:
        from onnxmltools.convert.lightgbm.operator_converters.LightGbm import (
            convert_lightgbm,
        )
        from skl2onnx.common.shape_calculator import (
            calculate_linear_classifier_output_shapes,
        )

        skl2onnx.update_registered_converter(
            lightgbm.LGBMClassifier,
            "LightGbmLGBMClassifier",
            calculate_linear_classifier_output_shapes,
            convert_lightgbm,
            options={"nocl": [True, False], "zipmap": [True, False]},
        )
    elif lightgbm.sklearn.LGBMRegressor in model_est_types:

        from onnxmltools.convert.lightgbm.operator_converters.LightGbm import (
            convert_lightgbm,
        )
        from skl2onnx.common.shape_calculator import (
            calculate_linear_regressor_output_shapes,
        )

        skl2onnx.update_registered_converter(
            lightgbm.LGBMRegressor,
            "LightGbmLGBMRegressor",
            calculate_linear_regressor_output_shapes,
            convert_lightgbm,
            options={"nocl": [True, False], "zipmap": [True, False]},
        )
    else:
        return onnxmltools.convert_lightgbm(
            model.est,
            name=None,
            initial_types=in_types,
            doc_string="",
            target_opset=None,
            targeted_onnx=onnx.__version__,
            custom_conversion_functions=None,
            custom_shape_calculators=None,
        )
    return skl2onnx.convert_sklearn(model.est, initial_types=in_types)


@runtime_dependency(module="xgboost", install_from=OptionalDependency.BOOSTED)
@runtime_dependency(module="skl2onnx", install_from=OptionalDependency.ONNX)
@runtime_dependency(module="onnxmltools", install_from=OptionalDependency.ONNX)
@runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
def _xgboost_to_onnx(model=None, target_dir=None, X=None, y=None, **kwargs):
    from skl2onnx.common.data_types import FloatTensorType

    in_types = [("input", FloatTensorType([None, len(X.columns)]))]

    if str(type(model.est)) == "<class 'sklearn.pipeline.Pipeline'>":
        model_est_types = [type(val[1]) for val in model.est.steps]
    else:
        model_est_types = [type(model.est)]
    if xgboost.sklearn.XGBClassifier in model_est_types:

        from onnxmltools.convert.xgboost.operator_converters.XGBoost import (
            convert_xgboost,
        )
        from skl2onnx.common.shape_calculator import (
            calculate_linear_classifier_output_shapes,
        )

        skl2onnx.update_registered_converter(
            xgboost.XGBClassifier,
            "XGBoostXGBClassifier",
            calculate_linear_classifier_output_shapes,
            convert_xgboost,
            options={"nocl": [True, False], "zipmap": [True, False]},
        )
    elif xgboost.sklearn.XGBRegressor in model_est_types:
        from onnxmltools.convert.xgboost.operator_converters.XGBoost import (
            convert_xgboost,
        )
        from skl2onnx.common.shape_calculator import (
            calculate_linear_regressor_output_shapes,
        )

        skl2onnx.update_registered_converter(
            xgboost.XGBRegressor,
            "XGBoostXGBRegressor",
            calculate_linear_regressor_output_shapes,
            convert_xgboost,
            options={"nocl": [True, False], "zipmap": [True, False]},
        )
    else:
        from onnxmltools.convert import convert_xgboost

        return convert_xgboost(
            model,
            name=None,
            initial_types=in_types,
            doc_string="",
            target_opset=None,
            targeted_onnx=onnx.__version__,
            custom_conversion_functions=None,
            custom_shape_calculators=None,
        )
    return skl2onnx.convert_sklearn(model.est, initial_types=in_types)


@runtime_dependency(module="torch", install_from=OptionalDependency.PYTORCH)
@runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
def _torch_to_onnx(model=None, target_dir=None, X=None, y=None, **kwargs):
    import torch

    assert hasattr(torch, "onnx"), (
        f"This version of pytorch {torch.__version__} does not appear to support onnx "
        f"conversion "
    )
    # Add variable batch size to the beginning of the shape
    sample_input = torch.randn([1] + [X[:1].shape[1]], requires_grad=True)
    torch.onnx.export(
        model.est,
        sample_input,
        f=os.path.join(target_dir, "model.onnx"),
        input_names=["input"],
        output_names=["output"],
        dynamic_axes={"input": {0: "batch_size"}, "output": {0: "batch_size"}},
    )


@runtime_dependency(module="onnxmltools", install_from=OptionalDependency.ONNX)
def _tf_to_onnx(model=None, target_dir=None, X=None, y=None, **kwargs):
    try:
        from onnxmltools.convert import convert_tensorflow

        return convert_tensorflow(
            model,
            name=None,
            input_names=None,
            output_names=None,
            doc_string="",
            target_opset=None,
            channel_first_inputs=None,
            debug_mode=False,
            custom_op_conversions=None,
        )
    except Exception as e:
        _log_automatic_serialization_keras(e)


@runtime_dependency(module="onnxmltools", install_from=OptionalDependency.ONNX)
@runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
def _keras_to_onnx(model=None, target_dir=None, X=None, y=None, **kwargs):
    try:
        from onnxmltools.convert import convert_keras

        return convert_keras(
            model,
            name=None,
            initial_types=None,
            doc_string="",
            target_opset=None,
            targeted_onnx=onnx.__version__,
            channel_first_inputs=None,
            custom_conversion_functions=None,
            custom_shape_calculators=None,
            default_batch_size=1,
        )
    except Exception as e:
        _log_automatic_serialization_keras(e)


@runtime_dependency(module="onnx", install_from=OptionalDependency.ONNX)
def _mxnet_to_onnx(model=None, target_dir=None, X=None, y=None, **kwargs):
    assert onnx.__version__ == "1.3.0", "Mxnet can only export to onnx version 1.3.0"
    from mxnet.contrib import onnx as onnx_mxnet

    sym = kwargs.get("sym", None)
    params = kwargs.get("params", None)
    assert sym is not None and params is not None, (
        "Pass `sym` [Path to the json file or Symbol object] and "
        "`params` [Path to the params file or params dictionary. "
        "(Including both arg_params and aux_params)] to kwargs"
    )
    input_shape = X.shape
    onnx_file = "model.onnx"
    onnx_mxnet.export_model(
        sym=sym,
        params=params,
        input_shape=[input_shape],
        input_type=np.float32,
        onnx_file_path=onnx_file,
    )


def _log_automatic_serialization_keras(e):
    logger.error(
        f"The following error occured: {e}. "
        f"This may be because automatic serialization for Keras models is not supported. "
        f"Use `prepare_generic_model()` method to manually serialize the model."
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import os
import tempfile
from typing import Any, Dict, Optional
from zipfile import ZipFile

import yaml

from ads.common import utils


DEPRECATE_AS_ONNX_WARNING = "This attribute `as_onnx` will be deprecated in the future. You can choose specific format by setting `model_save_serializer`."
DEPRECATE_USE_TORCH_SCRIPT_WARNING = "This attribute `use_torch_script` will be deprecated in the future. You can choose specific format by setting `model_save_serializer`."


def _extract_locals(
    locals: Dict[str, Any], filter_out_nulls: Optional[bool] = True
) -> Dict[str, Any]:
    """Extract arguments from local variables.
    If input dictionary contains `kwargs`, then it will not be included to
    the result dictionary.

    Properties
    ----------
    locals: Dict[str, Any]
        A dictionary, the result of `locals()` method.
        However can be any dictionary.
    filter_out_nulls: (bool, optional). Defaults to `True`.
        Whether `None` values should be filtered out from the result dict or not.

    Returns
    -------
    Dict[str, Any]
        A new dictionary with the result values.
    """
    result = {}
    keys_to_filter_out = ("kwargs",)
    consolidated_dict = {**locals.get("kwargs", {}), **locals}
    for key, value in consolidated_dict.items():
        if key not in keys_to_filter_out and not (filter_out_nulls and value is None):
            result[key] = value
    return result


def _is_json_serializable(data: Any) -> bool:
    """Check is data input is json serialization.

    Parameters
    ----------
    data: (Any)
        data to be passed to model for prediction.

    Returns
    -------
    bool
        Whether data is json serializable.
    """
    result = True
    try:
        json.dumps(data)
    except:
        result = False
    return result


def fetch_manifest_from_conda_location(env_location: str):
    """
    Convenience method to fetch the manifest file from a conda environment.

    :param env_location: Absolute path to the environment.
    :type env_location: str
    """
    manifest_location = None
    for file in os.listdir(env_location):
        if file.endswith("_manifest.yaml"):
            manifest_location = f"{env_location}/{file}"
            break
    env = {}
    if not manifest_location:
        raise Exception(
            f"Could not locate manifest file in the provided conda environment: {env_location}. Dir Listing - "
            f"{os.listdir(env_location)}"
        )

    with open(manifest_location) as mlf:
        env = yaml.load(mlf, Loader=yaml.FullLoader)
    manifest = env["manifest"]
    return manifest


def zip_artifact(artifact_dir: str) -> str:
    """Prepares model artifacts ZIP archive.

    Parameters
    ----------
    artifact_dir: str
        Path to the model artifact.

    Returns
    -------
    str
        Path to the model artifact ZIP archive file.
    """
    if not artifact_dir:
        raise ValueError("The `artifact_dir` must be provided.")
    if not os.path.exists(artifact_dir):
        raise ValueError(f"The {artifact_dir} not exists.")
    if not os.path.isdir(artifact_dir):
        raise ValueError("The `artifact_dir` must be a folder.")

    files_to_upload = utils.get_files(artifact_dir)
    # Set delete=False when creating NamedTemporaryFile,
    # Otherwise, the file will be delete when download_artifact() close the file.
    artifact = tempfile.NamedTemporaryFile(suffix=".zip", delete=False)
    # Close the file since NamedTemporaryFile() opens the file by default.
    artifact.close()
    artifact_zip_path = artifact.name

    with ZipFile(artifact_zip_path, "w") as zf:
        for matched_file in files_to_upload:
            zf.write(
                os.path.join(artifact_dir, matched_file),
                arcname=matched_file,
            )

    return artifact_zip_path



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ads
import os
import tempfile
import zipfile
from time import time
from typing import Dict, Optional, List, Union, Iterator

import pandas as pd
import numpy as np

import logging

logger = logging.getLogger("ads.msql_connector")


from mysql.connector import connection

CURSOR_SIZE = 50000


class MySQLRDBMSConnection(connection.MySQLConnection):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if kwargs["user_name"]:
            kwargs["user"] = kwargs.pop("user_name")
        super().__init__(*args, **kwargs)

    def insert(
        self,
        table_name: str,
        df: pd.DataFrame,
        if_exists: str,
        batch_size=100000,
        encoding="utf-8",
    ):

        if if_exists not in ["fail", "replace", "append"]:
            raise ValueError(
                f"Unknown option `if_exists`={if_exists}. Valid options are 'fail', 'replace', 'append'"
            )
        start_time = time()

        df_orcl = df.copy()
        # "object" type can be actual objects, or just plain strings, when inserting into the
        # database we need to stringify these so they can be represented in a VARCHAR2 column

        # object_columns = df_orcl.select_dtypes(include=["object"]).columns
        # df_orcl = df_orcl.where(pd.notnull(df_orcl), None)
        # df_orcl[object_columns] = df_orcl[object_columns].astype(str)

        # prep column names for valid Oracle column names (alpha + # $ _)
        df_orcl.columns = df_orcl.columns.str.replace(r"\W+", "_", regex=True)
        table_exist = True
        with self.cursor() as cursor:

            if if_exists != "replace":
                try:
                    cursor.execute(f"SELECT 1 from {table_name} LIMIT 1")
                    cursor.fetchall()
                except Exception as e:
                    table_exist = False
                if if_exists == "fail" and table_exist:
                    raise ValueError(
                        f"Table {table_name} already exists. Set `if_exists`='replace' or 'append' to replace or append to the existing table"
                    )
            # Oracle doesn't have boolean so convert to 1/0
            df_orcl = df_orcl.where(
                df.applymap(type) != bool, df_orcl.replace({True: 1, False: 0})
            )

            type_mappings = {
                "bool": "NUMBER(1)",
                "int16": "INTEGER",
                "int32": "INTEGER",
                "int64": "INTEGER",
                "float16": "FLOAT",
                "float32": "FLOAT",
                "float64": "FLOAT",
                "datetime64": "DATETIME",
            }
            # add in any string types as Oracle's VARCHAR type setting length to accommodate longest
            def get_max_str_len(df, column, encoding):
                return df[column].dropna().str.encode(encoding).map(len).max()

            longest_string_column = max(
                (
                    get_max_str_len(df_orcl, c, encoding)
                    for c in df_orcl.select_dtypes(
                        include=["object", "category"]
                    ).columns
                ),
                default=0,
            )

            logger.debug(f"Max string column value: {longest_string_column}")

            datatypes = {
                c: f"VARCHAR({get_max_str_len(df_orcl, c, encoding)})"
                for c in df_orcl.select_dtypes(include=["object", "category"]).columns
            }

            for df_type, orcl_type in type_mappings.items():

                datatypes.update(
                    {
                        column: orcl_type
                        for column in df_orcl.select_dtypes(include=df_type).columns
                    }
                )

            if set(datatypes.keys()) != set(df_orcl.columns):
                raise Exception(
                    f"Unable to determine MySQL data type to use for column(s): {', '.join(set(df_orcl.columns)-set(datatypes.keys()))}"
                )

            # create table
            if if_exists == "replace":
                try:
                    cursor.execute(f"drop table {table_name}")
                except:
                    logger.info(f"Table {table_name} does not exist")
            if if_exists == "replace" or not table_exist:
                sql = (
                    f"create table {table_name} ("
                    + ", ".join([f"{col} {datatypes[col]}" for col in df_orcl.columns])
                    + ")"
                )
                logger.info(sql)
                try:
                    cursor.execute(sql)
                except Exception as e:
                    raise Exception(
                        f"'{e if e.args else 'Unexpected error encountered'}' with query: '{sql}'"
                    ) from e

            # insert
            bind_columns = ", ".join([f"{col}" for col in df_orcl.columns])
            bind_variables = ",".join(["%s" for _ in df_orcl.columns])
            sql = f"insert into {table_name}({bind_columns}) values({bind_variables})"

            logger.info(sql)

            # prevent buffer reallocation by locking in the longest string value
            # cursor.setinputsizes(None, longest_string_column)

            # replace NaN with None before turning into database records, important - don't
            # do this earlier in the logic because it can change the pandas
            # data types

            record_data = list(
                (
                    [
                        xr.to_pydatetime().strftime("%Y-%m-%d %H:%M:%S")
                        if hasattr(xr, "to_pydatetime")
                        else xr
                        for xr in x
                    ]
                    for x in df_orcl.replace({np.nan: None}).itertuples(
                        index=False, name=None
                    )
                )
            )

            def chunks(lst: List, batch_size: int):
                """Yield successive batch_size chunks from lst."""
                for i in range(0, len(lst), batch_size):
                    yield lst[i : i + batch_size]

            for batch in chunks(record_data, batch_size=batch_size):
                cursor.executemany(sql, batch)

            self.commit()

            duration = time() - start_time
            logger.info(
                f"inserted {df_orcl.shape[0]} rows at {df_orcl.shape[0]/duration:.2f} rows/seconds"
            )

    def _fetch_by_batch(self, cursor, chunksize):
        while True:
            rows = cursor.fetchmany(chunksize)
            if rows:
                yield rows
            else:
                break

    def query(
        self, sql: str, bind_variables: Optional[Dict], chunksize=None
    ) -> Union[pd.DataFrame, Iterator[pd.DataFrame]]:

        start_time = time()

        cursor = self.cursor(prepared=True)
        cursor.arraysize = CURSOR_SIZE

        cursor.execute(sql, bind_variables)
        columns = [row[0] for row in cursor.description]

        if chunksize:
            logger.info(f"Chunksize is {chunksize}")
            df = iter(
                (
                    pd.DataFrame(data=rows, columns=columns)
                    for rows in self._fetch_by_batch(cursor, chunksize)
                )
            )

        else:
            df = pd.DataFrame(
                cursor,
                columns=columns,
            )
            duration = time() - start_time
            logger.info(
                f"fetched {df.shape[0]} rows at {df.shape[0]/duration:.2f} rows/seconds"
            )

        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import dataclasses
import json
import os
from typing import Any, Dict, Optional, Union

from ads.common.config import Config, ConfigSection
from ads.common.serializer import Serializable


@dataclasses.dataclass(repr=False)
class BaseProperties(Serializable):
    """Represents base properties class.

    Methods
    -------
    with_prop(name: str, value: Any) -> BaseProperties
        Sets property value.
    with_dict(obj_dict: Dict) -> BaseProperties
        Populates properties values from dict.
    with_env() -> BaseProperties
        Populates properties values from environment variables.
    to_dict() -> Dict
        Serializes instance of class into a dictionary.
    with_config(config: ads.config.ConfigSection) -> BaseProperties
        Sets properties values from the config profile.
    from_dict(obj_dict: Dict[str, Any]) -> "BaseProperties"
        Creates an instance of the properties class from a dictionary.
    from_config(uri: str, profile: str, auth: Optional[Dict] = None) -> "BaseProperties":
        Loads properties from the config file.
    to_config(uri: str, profile: str, force_overwrite: Optional[bool] = False, auth: Optional[Dict] = None) -> None
        Saves properties to the config file.
    """

    def __setattr__(self, name: str, value: Any):
        """Adds type validation when attribute assignment is attempted."""
        if value is None:
            self.__dict__[name] = value
        elif name in self.__annotations__:
            if hasattr(self.__annotations__[name], "__origin__"):
                if self.__annotations__[name].__origin__ is Union and not isinstance(
                    value, self.__annotations__[name].__args__
                ):
                    raise TypeError(
                        f"Field `{name}` was expected to be of type `{self.__annotations__[name].__args__}` "
                        f"but type `{type(value)}` was provided."
                    )

            elif self.__annotations__[name] != type(value):
                raise TypeError(
                    f"Field `{name}` was expected to be of type `{self.__annotations__[name]}` "
                    f"but type `{type(value)}` was provided."
                )

            self.__dict__[name] = value

    def with_prop(self, name: str, value: Any) -> "BaseProperties":
        """Sets property value.

        Parameters
        ----------
        name: str
            Property name.
        value:
            Property value.

        Returns
        -------
        BaseProperties
            Instance of the BaseProperties.
        """
        setattr(self, name, value)
        return self

    def with_dict(self, obj_dict: Dict[str, Any]) -> "BaseProperties":
        """Sets properties from a dict.

        Parameters
        ----------
        obj_dict: Dict[str, Any]
            List of properties and values in dictionary format.

        Returns
        -------
        BaseProperties
            Instance of the BaseProperties.

        Raises
        ------
        TypeError
            If input object has a wrong type.
        """
        if obj_dict is None:
            return self

        if not isinstance(obj_dict, Dict):
            raise TypeError("The `obj_dict` should be a dictionary.")

        for key, value in obj_dict.items():
            # if expected type of input value is not a string, but
            # actual type of the value is string, then try to convert value to the
            # expected format by using JSON.loads()
            if (
                not value is None
                and key in self.__annotations__
                and self.__annotations__[key] != str
                and isinstance(value, str)
            ):
                try:
                    value = json.loads(value)
                except:
                    pass

            self.with_prop(key, value)
        return self

    @classmethod
    def from_dict(cls, obj_dict: Dict[str, Any]) -> "BaseProperties":
        """Creates an instance of the properties class from a dictionary.

        Parameters
        ----------
        obj_dict: Dict[str, Any]
            List of properties and values in dictionary format.

        Returns
        -------
        BaseProperties
            Instance of the BaseProperties.
        """
        return cls().with_dict(obj_dict)

    def with_env(self) -> "BaseProperties":
        """Sets properties values from environment variables.

        Returns
        -------
        BaseProperties
            Instance of the BaseProperties.
        """
        self.with_dict(
            {env_k.lower(): env_v for env_k, env_v in sorted(os.environ.items())}
        )
        self._adjust_with_env()
        return self

    def with_config(self, config: ConfigSection) -> "BaseProperties":
        """Sets properties values from the config profile.

        Returns
        -------
        BaseProperties
            Instance of the BaseProperties.
        """
        if not isinstance(config, ConfigSection):
            raise TypeError("`config` should be an instance of `ConfigSection`.")

        return self.with_dict(config.to_dict())

    def to_dict(self, **kwargs):
        """Serializes instance of class into a dictionary.

        Returns
        -------
        Dict
            A dictionary.
        """
        return dataclasses.asdict(self)

    def to_config(
        self,
        uri: str,
        profile: str,
        force_overwrite: Optional[bool] = False,
        auth: Optional[Dict] = None,
    ) -> None:
        """Saves properties to the config file.

        Parameters
        ----------
        uri: str
            The URI of the config file.
            Can be local path or OCI object storage URI.
        profile: str
            The config profile name.
        force_overwrite: (bool, optional). Defaults to False.
            Whether to overwrite existing files or not.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        None
            Nothing
        """
        config = Config(uri=uri, auth=auth)
        config.section_set(key=profile, info=self.to_dict(), replace=force_overwrite)
        config.save()

    @classmethod
    def from_config(
        cls,
        uri: str,
        profile: str,
        auth: Optional[Dict] = None,
    ) -> "BaseProperties":
        """Loads properties from the config file.

        Parameters
        ----------
        uri: str
            The URI of the config file.
            Can be local path or OCI object storage URI.
        profile: str
            The config profile name.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        BaseProperties
            Instance of the BaseProperties.
        """
        config = Config(uri=uri, auth=auth).load()
        return cls().with_config(config[profile])

    def _adjust_with_env(self) -> None:
        """Adjusts env variables. The method is used in `with_env` method."""
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.model.extractor.xgboost_extractor import XgboostExtractor
from ads.model.generic_model import FrameworkSpecificModel
from ads.model.model_properties import ModelProperties
from ads.model.serde.model_serializer import XgboostModelSerializerType
from ads.model.common.utils import DEPRECATE_AS_ONNX_WARNING
from ads.model.serde.common import SERDE


class XGBoostModel(FrameworkSpecificModel):
    """XGBoostModel class for estimators from xgboost framework.

    Attributes
    ----------
    algorithm: str
        The algorithm of the model.
    artifact_dir: str
        Artifact directory to store the files needed for deployment.
    auth: Dict
        Default authentication is set using the `ads.set_auth` API. To override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create
        an authentication signer to instantiate an IdentityClient object.
    estimator: Callable
        A trained xgboost estimator/model using Xgboost.
    framework: str
        "xgboost", the framework name of the model.
    hyperparameter: dict
        The hyperparameters of the estimator.
    metadata_custom: ModelCustomMetadata
        The model custom metadata.
    metadata_provenance: ModelProvenanceMetadata
        The model provenance metadata.
    metadata_taxonomy: ModelTaxonomyMetadata
        The model taxonomy metadata.
    model_artifact: ModelArtifact
        This is built by calling prepare.
    model_deployment: ModelDeployment
        A ModelDeployment instance.
    model_file_name: str
        Name of the serialized model.
    model_id: str
        The model ID.
    properties: ModelProperties
        ModelProperties object required to save and deploy model.
        For more details, check https://accelerated-data-science.readthedocs.io/en/latest/ads.model.html#module-ads.model.model_properties.
    runtime_info: RuntimeInfo
        A RuntimeInfo instance.
    schema_input: Schema
        Schema describes the structure of the input data.
    schema_output: Schema
        Schema describes the structure of the output data.
    serialize: bool
        Whether to serialize the model to pkl file by default. If False, you need to serialize the model manually,
        save it under artifact_dir and update the score.py manually.
    version: str
        The framework version of the model.

    Methods
    -------
    delete_deployment(...)
        Deletes the current model deployment.
    deploy(..., **kwargs)
        Deploys a model.
    from_model_artifact(uri, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from the specified folder, or zip/tar archive.
    from_model_catalog(model_id, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from model catalog.
    introspect(...)
        Runs model introspection.
    predict(data, ...)
        Returns prediction of input data run against the model deployment endpoint.
    prepare(..., **kwargs)
        Prepare and save the score.py, serialized model and runtime.yaml file.
    reload(...)
        Reloads the model artifact files: `score.py` and the `runtime.yaml`.
    save(..., **kwargs)
        Saves model artifacts to the model catalog.
    summary_status(...)
        Gets a summary table of the current status.
    verify(data, ...)
        Tests if deployment works in local environment.

    Examples
    --------
    >>> import xgboost as xgb
    >>> import tempfile
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.datasets import load_iris
    >>> from ads.model.framework.xgboost_model import XGBoostModel

    >>> iris = load_iris()
    >>> X, y = iris.data, iris.target
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25)
    >>> xgboost_estimator = xgb.XGBClassifier()
    >>> xgboost_estimator.fit(X_train, y_train)

    >>> xgboost_model = XGBoostModel(estimator=xgboost_estimator, artifact_dir=tmp_model_dir)
    >>> xgboost_model.prepare(inference_conda_env="generalml_p37_cpu_v1", force_overwrite=True)
    >>> xgboost_model.reload()
    >>> xgboost_model.verify(X_test)
    >>> xgboost_model.save()
    >>> model_deployment = xgboost_model.deploy(wait_for_completion=False)
    >>> xgboost_model.predict(X_test)
    """

    _PREFIX = "xgboost"
    model_save_serializer_type = XgboostModelSerializerType

    @runtime_dependency(module="xgboost", install_from=OptionalDependency.BOOSTED)
    def __init__(
        self,
        estimator: callable,
        artifact_dir: str,
        properties: Optional[ModelProperties] = None,
        auth: Dict = None,
        model_save_serializer: Optional[SERDE] = model_save_serializer_type.XGBOOST,
        model_input_serializer: Optional[SERDE] = None,
        **kwargs,
    ):
        """
        Initiates a XGBoostModel instance. This class wraps the XGBoost model as estimator.
        It's primary purpose is to hold the trained model and do serialization.

        Parameters
        ----------
        estimator:
            XGBoostModel
        artifact_dir: str
            artifact directory to store the files needed for deployment.
        properties: (ModelProperties, optional). Defaults to None.
            ModelProperties object required to save and deploy model.
        auth :(Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        model_save_serializer: (SERDE or str, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize model.
        model_input_serializer: (SERDE, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize data.

        Returns
        -------
        XGBoostModel
            XGBoostModel instance.


        Examples
        --------
        >>> import xgboost as xgb
        >>> import tempfile
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.model_selection import train_test_split
        >>> from sklearn.datasets import load_iris
        >>> from ads.model.framework.xgboost_model import XGBoostModel

        >>> iris = load_iris()
        >>> X, y = iris.data, iris.target

        >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25)
        >>> train = xgb.DMatrix(X_train, y_train)
        >>> test = xgb.DMatrix(X_test, y_test)
        >>> xgboost_estimator = XGBClassifier()
        >>> xgboost_estimator.fit(X_train, y_train)
        >>> xgboost_model = XGBoostModel(estimator=xgboost_estimator, artifact_dir=tempfile.mkdtemp())
        >>> xgboost_model.prepare(inference_conda_env="generalml_p37_cpu_v1")
        >>> xgboost_model.verify(X_test)
        >>> xgboost_model.save()
        >>> model_deployment = xgboost_model.deploy()
        >>> xgboost_model.predict(X_test)
        >>> xgboost_model.delete_deployment()
        """
        if not (
            str(type(estimator)).startswith("<class 'xgboost.")
            or str(type(estimator)).startswith("<class 'onnxruntime.")
        ):
            raise TypeError(f"{str(type(estimator))} is not supported in XGBoostModel.")
        super().__init__(
            estimator=estimator,
            artifact_dir=artifact_dir,
            properties=properties,
            auth=auth,
            model_save_serializer=model_save_serializer,
            model_input_serializer=model_input_serializer,
            **kwargs,
        )
        self._extractor = XgboostExtractor(estimator)
        self.framework = self._extractor.framework
        self.algorithm = self._extractor.algorithm
        self.version = self._extractor.version
        self.hyperparameter = self._extractor.hyperparameter

    @runtime_dependency(module="xgboost", install_from=OptionalDependency.BOOSTED)
    def serialize_model(
        self,
        as_onnx: bool = False,
        initial_types: List[Tuple] = None,
        force_overwrite: bool = False,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        **kwargs,
    ):
        """
        Serialize and save Xgboost model using ONNX or model specific method.

        Parameters
        ----------
        artifact_dir: str
            Directory for generate artifact.
        as_onnx: (boolean, optional). Defaults to False.
            If set as True, provide initial_types or X_sample to convert into ONNX.
        initial_types: (List[Tuple], optional). Defaults to None.
            Each element is a tuple of a variable name and a type.
        force_overwrite: (boolean, optional). Defaults to False.
            If set as True, overwrite serialized model if exists.
        X_sample: Union[Dict, str, List, np.ndarray, pd.core.series.Series, pd.core.frame.DataFrame,]. Defaults to None.
            Contains model inputs such that model(X_sample) is a valid invocation of the model.
            Used to generate initial_types.

        Returns
        -------
        None
            Nothing.
        """
        if as_onnx:
            logger.warning(DEPRECATE_AS_ONNX_WARNING)
            self.set_model_save_serializer(self.model_save_serializer_type.ONNX)

        super().serialize_model(
            as_onnx=as_onnx,
            initial_types=initial_types,
            force_overwrite=force_overwrite,
            X_sample=X_sample,
            **kwargs,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a CreditCard feature type.

Classes:
    CreditCard
        The CreditCard feature type.

Functions:
    default_handler(data: pd.Series) -> pd.Series
        Processes given data and indicates if the data matches requirements.
    _luhn_checksum(card_number: str) -> float
        Implements Luhn algorithm to validate a credit card number.
"""
import matplotlib.pyplot as plt
import pandas as pd
import re
from ads.feature_engineering.feature_type.string import String
from ads.feature_engineering.utils import (
    assign_issuer,
    _count_unique_missing,
    _set_seaborn_theme,
    SchemeTeal,
)
from ads.feature_engineering import schema
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


_max_sample_size_to_luhn_check = 1000
_pattern_string = r"""
        ^(?:4[0-9]{12}(?:[0-9]{3})?         # Visa
        |  (?:5[1-5][0-9]{2}                # MasterCard
        | 222[1-9]|22[3-9][0-9]|2[3-6][0-9]{2}|27[01][0-9]|2720)[0-9]{12}
        |  3[47][0-9]{13}                   # American Express
        |  3(?:0[0-5]|[68][0-9])[0-9]{11}   # Diners Club
        |  6(?:011|5[0-9]{2})[0-9]{12}      # Discover
        |  (?:2131|1800|35\d{3})\d{11}      # JCB
        |  (5018|5020|5038|5612|5893|6304|6759|6761|6762|6763|0604|6390)\d+$   # Maestro
        |  ^(5[06789]|6)[0-9]{0,}$          # Maestro
        |  ^4[0-9]{12}(?:[0-9]{6})?$        #Visa 19 digit
        )$
    """
PATTERN = re.compile(_pattern_string, re.VERBOSE)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """
    Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: :class:`pandas.Series`
        The data to process.

    Returns
    -------
    :class:`pandas.Series`
        The logical list indicating if the data matches requirements.
    """

    def _is_credit_card(x: pd.Series):
        return (
            not pd.isnull(x)
            and PATTERN.match(str(x)) is not None
            and _luhn_checksum(str(x)) == 0
        )

    return data.apply(lambda x: True if _is_credit_card(x) else False)


def _luhn_checksum(card_number: str) -> float:
    """
    Implements Luhn algorithm to validate a credit card number.

    Parameters
    ----------
    card_number : str
        The credit card number.

    Returns
    -------
    float
        The checksum of the card number
    """

    def digits_of(n):
        return [int(d) for d in str(n)]

    digits = digits_of(card_number)
    odd_digits = digits[-1::-2]
    even_digits = digits[-2::-2]
    checksum = 0
    checksum += sum(odd_digits)
    for d in even_digits:
        checksum += sum(digits_of(d * 2))
    return checksum % 10


class CreditCard(String):
    """
    Type representing credit card numbers.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows the counts of observations in each credit card type using bar chart.


    Examples
    --------
    >>> from ads.feature_engineering.feature_type.creditcard import CreditCard
    >>> import pandas as pd
    >>> s = pd.Series(["4532640527811543", None, "4556929308150929", "4539944650919740", "4485348152450846", "4556593717607190"], name='credit_card')
    >>> s.ads.feature_type = ['credit_card']
    >>> CreditCard.validator.is_credit_card(s)
    0     True
    1    False
    2     True
    3     True
    4     True
    5     True
    Name: credit_card, dtype: bool
    """

    description = "Type representing credit card numbers."

    @staticmethod
    def feature_stat(x: pd.Series):
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count), missing(count) and
            count of each credit card type.

        Examples
        --------
        >>> visa = [
            "4532640527811543",
            None,
            "4556929308150929",
            "4539944650919740",
            "4485348152450846",
            "4556593717607190",
            ]
        >>> mastercard = [
            "5334180299390324",
            "5111466404826446",
            "5273114895302717",
            "5430972152222336",
            "5536426859893306",
            ]
        >>> amex = [
            "371025944923273",
            "374745112042294",
            "340984902710890",
            "375767928645325",
            "370720852891659",
            ]
        >>> creditcard_list = visa + mastercard + amex
        >>> creditcard_series = pd.Series(creditcard_list,name='card')
        >>> creditcard_series.ads.feature_type = ['credit_card']
        >>> creditcard_series.ads.feature_stat()
            Metric              Value
        0	count	            16
        1	unique	            15
        2	missing	            1
        3	count_Amex	        5
        4	count_Visa	        5
        5	count_MasterCard	3
        6	count_Diners Club	2
        7	count_missing	    1

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series or Dataframe provided.
        """
        df_stat = _count_unique_missing(x)
        card_types = x.apply(assign_issuer)
        value_counts = card_types.value_counts()
        value_counts.index = [
            "count_" + cardtype for cardtype in list(value_counts.index)
        ]
        return pd.concat([df_stat, value_counts.to_frame()])

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows the counts of observations in each credit card type using bar chart.

        Examples
        --------
        >>> visa = [
            "4532640527811543",
            None,
            "4556929308150929",
            "4539944650919740",
            "4485348152450846",
            "4556593717607190",
            ]
        >>> mastercard = [
            "5334180299390324",
            "5111466404826446",
            "5273114895302717",
            "5430972152222336",
            "5536426859893306",
            ]
        >>> amex = [
            "371025944923273",
            "374745112042294",
            "340984902710890",
            "375767928645325",
            "370720852891659",
            ]
        >>> creditcard_list = visa + mastercard + amex
        >>> creditcard_series = pd.Series(creditcard_list,name='card')
        >>> creditcard_series.ads.feature_type = ['credit_card']
        >>> creditcard_series.ads.feature_plot()

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the CreditCard feature type.
        """
        card_types = x.apply(assign_issuer)
        df = card_types.value_counts().to_frame()
        if len(df.index):
            _set_seaborn_theme()
            ax = seaborn.barplot(
                y=df.index, x=list(df.iloc[:, 0]), color=SchemeTeal.AREA_DARK
            )
            ax.set(xlabel="Count")
            return ax

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> visa = [
            "4532640527811543",
            None,
            "4556929308150929",
            "4539944650919740",
            "4485348152450846",
            "4556593717607190",
            ]
        >>> mastercard = [
            "5334180299390324",
            "5111466404826446",
            "5273114895302717",
            "5430972152222336",
            "5536426859893306",
            ]
        >>> amex = [
            "371025944923273",
            "374745112042294",
            "340984902710890",
            "375767928645325",
            "370720852891659",
            ]
        >>> creditcard_list = visa + mastercard + amex
        >>> creditcard_series = pd.Series(creditcard_list,name='card')
        >>> creditcard_series.ads.feature_type = ['credit_card']
        >>> creditcard_series.ads.feature_domain()
        constraints: []
        stats:
            count: 16
            count_Amex: 5
            count_Diners Club: 2
            count_MasterCard: 3
            count_Visa: 5
            count_missing: 1
            missing: 1
            unique: 15
        values: CreditCard

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the CreditCard feature type.
        """
        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [],
        )


CreditCard.validator.register("is_credit_card", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import datetime
import os
import shutil
import subprocess
import shlex
import json
import glob
from typing import Dict

import click
import yaml

from datetime import datetime
import ocifs

from ads.common.oci_client import OCIClientFactory
from ads.common.auth import create_signer
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.config import NO_CONTAINER

from ads.opctl.constants import (
    ML_JOB_GPU_IMAGE,
    ML_JOB_IMAGE,
    DEFAULT_IMAGE_HOME_DIR,
    DEFAULT_IMAGE_CONDA_DIR,
    DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR,
)
from ads.opctl import logger
from ads.opctl.config.utils import read_from_ini
from ads.opctl.utils import (
    parse_conda_uri,
    run_container,
    get_docker_client,
    is_in_notebook_session,
    run_command,
)
from ads.opctl.config.base import ConfigProcessor
from ads.opctl.config.merger import ConfigMerger
from ads.opctl.conda.multipart_uploader import MultiPartUploader


def _fetch_manifest_template() -> Dict:
    manifest_template_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "manifest_template.yaml"
    )
    with open(manifest_template_file) as manifest_file:
        manifest_template = yaml.safe_load(manifest_file)
    return manifest_template


@runtime_dependency(module="docker", install_from=OptionalDependency.OPCTL)
def _check_job_image_exists(gpu: bool) -> None:
    if gpu:
        image = ML_JOB_GPU_IMAGE
    else:
        image = ML_JOB_IMAGE
    try:
        client = get_docker_client()
        client.api.inspect_image(image)
    except docker.errors.ImageNotFound:
        if gpu:
            cmd = "`ads opctl build-image -g job-local`"
        else:
            cmd = "`ads opctl build-image job-local`"
        raise RuntimeError(
            f"Please run {cmd} to build a local image for Jobs development first."
        )


def _get_name(name: str, env_file: str) -> str:
    if not name and env_file:
        with open(env_file) as f:
            name = yaml.safe_load(f.read()).get("name", None)
    if not name:
        raise ValueError(
            "Either specify environment name in environment yaml or with `--name`."
        )
    return name


def create(**kwargs) -> str:
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    exec_config = p.config["execution"]
    name = _get_name(exec_config.get("name"), exec_config.get("environment_file"))
    _create(
        name=name,
        version=exec_config["version"],
        env_file=exec_config["environment_file"],
        conda_pack_folder=exec_config["conda_pack_folder"],
        gpu=exec_config.get("gpu", False),
        overwrite=exec_config.get("overwrite", False),
    )


def _create(
    name: str,
    version: str,
    env_file: str,
    conda_pack_folder: str,
    gpu: bool,
    overwrite: bool,
) -> str:
    """Create a conda pack given an environment yaml file under conda pack folder specified.

    Parameters
    ----------
    slug : str
        slug of the conda pack
    env_file : str
        path to conda environment yaml
    conda_pack_folder : str
        path to conda pack folder
    gpu : bool
        whether to build against GPU image
    overwrite : bool
        whether to overwrite existing pack of the same slug

    Raises
    ------
    FileNotFoundError
        Environment file not found
    RuntimeError
        Environment file may be inproperly removed
    RuntimeError
        Creating pack failed

    Return
    ------
    str:
        slug of the environment
    """

    if not os.path.exists(env_file):
        raise FileNotFoundError(f"Environment file {env_file} is not found.")

    slug = f"{name}_v{version}".replace(" ", "").replace(".", "_").lower()
    pack_folder_path = os.path.join(
        os.path.abspath(os.path.expanduser(conda_pack_folder)), slug
    )

    if os.path.exists(pack_folder_path):
        overwrite = overwrite or click.confirm(
            f"Conda pack with slug {slug} already exists. Do you wish to overwrite?"
        )
        if overwrite:
            if (
                os.path.commonpath(
                    [pack_folder_path, os.path.abspath(os.path.expanduser(env_file))]
                )
                == pack_folder_path
            ):
                raise RuntimeError(
                    f"Environment file {os.path.abspath(os.path.expanduser(env_file))} is in {pack_folder_path}. Overwriting this folder will remove the file."
                )
            shutil.rmtree(pack_folder_path)
        else:
            return

    os.makedirs(pack_folder_path, exist_ok=True)

    manifest = _fetch_manifest_template()
    manifest["manifest"]["name"] = name
    manifest["manifest"]["slug"] = slug
    manifest["manifest"]["type"] = "published"
    manifest["manifest"]["version"] = version
    manifest["manifest"]["arch_type"] = "GPU" if gpu else "CPU"

    manifest["manifest"]["create_date"] = datetime.utcnow().strftime(
        "%a, %b %d, %Y, %H:%M:%S %Z UTC"
    )
    manifest["manifest"]["manifest_version"] = "1.0"

    logger.info(f"Creating conda environment {slug}")
    if is_in_notebook_session() or NO_CONTAINER:
        command = f"conda env create --prefix {pack_folder_path} --file {os.path.abspath(os.path.expanduser(env_file))}"
        run_command(command, shell=True)
    else:
        _check_job_image_exists(gpu)
        docker_pack_folder_path = os.path.join(DEFAULT_IMAGE_HOME_DIR, slug)
        docker_env_file_path = os.path.join(
            DEFAULT_IMAGE_HOME_DIR, os.path.basename(env_file)
        )

        create_command = f"conda env create --prefix {docker_pack_folder_path} --file {docker_env_file_path}"

        volumes = {
            pack_folder_path: {"bind": docker_pack_folder_path},
            os.path.abspath(os.path.expanduser(env_file)): {
                "bind": docker_env_file_path
            },
        }
        if gpu:
            image = ML_JOB_GPU_IMAGE
        else:
            image = ML_JOB_IMAGE
        try:
            run_container(
                image=image, bind_volumes=volumes, env_vars={}, command=create_command
            )
        except Exception:
            if os.path.exists(pack_folder_path):
                shutil.rmtree(pack_folder_path)
            raise RuntimeError(f"Could not create environment {slug}.")

    conda_dep = None
    with open(env_file) as mfile:
        conda_dep = yaml.safe_load(mfile.read())
    conda_dep["manifest"] = manifest["manifest"]
    with open(f"{os.path.join(pack_folder_path, slug)}_manifest.yaml", "w") as mfile:
        yaml.safe_dump(conda_dep, mfile)

    logger.info(f"Environment `{slug}` setup complete.")
    print(f"Pack {slug} created under {pack_folder_path}.")
    return slug


def _fetch_pack_path(
    slug: str,
    conda_pack_os_prefix: str,
    oci_config: str,
    oci_profile: str,
    auth_type: str,
) -> str:
    oci_auth = create_signer(auth_type, oci_config, oci_profile)
    fs = ocifs.OCIFileSystem(**oci_auth)
    fnames = fs.ls(conda_pack_os_prefix, detail=True, refresh=True)
    for f in fnames:
        if os.path.basename(f["name"]) == slug:
            return f"oci://{f['name']}"
        elif f["type"] == "directory":
            path = _fetch_pack_path(
                slug, f"oci://{f['name']}", oci_config, oci_profile, auth_type
            )
            if path:
                return path
    return None


def install(**kwargs) -> None:
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    exec_config = p.config["execution"]
    if exec_config.get("slug") and exec_config.get("conda_pack_os_prefix"):
        conda_uri = _fetch_pack_path(
            exec_config["slug"],
            exec_config["conda_pack_os_prefix"],
            exec_config.get("oci_config"),
            exec_config.get("oci_profile"),
            exec_config.get("auth"),
        )
        if not conda_uri:
            raise FileNotFoundError(
                f"{exec_config['slug']} not found under {exec_config['conda_pack_os_prefix']}"
            )
    elif exec_config.get("conda_uri"):
        conda_uri = exec_config["conda_uri"]
    else:
        raise ValueError(
            "Either `--conda-uri`, or `--slug` and `--conda-pack-os-prefix` has to be specified."
        )
    _install(
        conda_uri,
        os.path.abspath(os.path.expanduser(exec_config["conda_pack_folder"])),
        exec_config.get("oci_config"),
        exec_config.get("oci_profile"),
        overwrite=exec_config.get("overwrite", False),
        debug=kwargs.get("debug", False),
        auth_type=exec_config.get("auth"),
    )


def _install(
    conda_uri: str,
    conda_pack_folder: str,
    oci_config: str = None,
    oci_profile: str = None,
    overwrite: bool = False,
    debug: bool = False,
    auth_type: str = None,
) -> None:
    """
    Install conda pack.

    Parameters
    ----------
    conda_uri: str
        OCI object storage uri to conda pack
    conda_pack_folder: str
        local folder to save conda packs
    oci_config: str
        path to OCI config file
    oci_profile: str
        OCI config profile
    overwrite: bool
        whether to overwrite existing pack
    debug: bool
        whether to turn on debug mode
    auth_type : str
        authentication method
    Returns
    -------
    None
    """
    ns, bucket, path, slug = parse_conda_uri(conda_uri)
    os.makedirs(conda_pack_folder, exist_ok=True)
    pack_path = os.path.join(os.path.expanduser(conda_pack_folder), slug + ".tar.gz")
    pack_folder_path = os.path.join(os.path.expanduser(conda_pack_folder), slug)

    if not (is_in_notebook_session() or NO_CONTAINER):
        _check_job_image_exists(gpu=False)

    while os.path.exists(pack_folder_path):
        if overwrite:
            break
        else:
            ans = click.prompt(
                f"conda pack with slug {slug} already exists. Enter a new name or 'o' for overwrite:",
                default="o",
            )
            if ans == "o":
                overwrite = True
                break
            else:
                slug = ans
                pack_path = os.path.join(
                    os.path.expanduser(conda_pack_folder), slug + ".tar.gz"
                )
                pack_folder_path = os.path.join(
                    os.path.expanduser(conda_pack_folder), slug
                )

    if oci_config is None or oci_profile is None:
        download_command = [
            "oci",
            "os",
            "object",
            "get",
            "--name",
            f"{path}",
            "-bn",
            bucket,
            "-ns",
            ns,
            "--file",
            pack_path,
            "--auth",
            auth_type,
        ]
    else:
        oci_config = os.path.abspath(os.path.expanduser(oci_config))
        parser = read_from_ini(os.path.expanduser(oci_config))
        if oci_profile not in parser:
            raise ValueError(f"PROFILE {oci_profile} not found in {oci_config}.")
        download_command = [
            "oci",
            "os",
            "object",
            "get",
            "--name",
            f"{path}",
            "-bn",
            bucket,
            "-ns",
            ns,
            "--file",
            pack_path,
            "--profile",
            oci_profile,
            "--config-file",
            oci_config,
        ]
    if debug:
        download_command.append("-d")
        print(" ".join([shlex.quote(c) for c in download_command]))
    process = subprocess.Popen(download_command)
    process.communicate()
    if process.returncode != 0:
        if os.path.exists(pack_path):
            os.remove(pack_path)
        raise RuntimeError(
            f"Downloading pack failed with return code {process.returncode}. Please double check the path and make sure you have access."
        )
    else:
        print(f"Download {conda_uri} completed")

    print(f"Start unpacking {pack_path}")
    os.makedirs(pack_folder_path, exist_ok=True)

    process = subprocess.Popen(["tar", "-xf", pack_path, "-C", pack_folder_path])
    process.communicate()
    if process.returncode != 0:
        shutil.rmtree(pack_folder_path)
        raise Exception(
            f"Error extracting {pack_path} to {pack_folder_path}. Please try again."
        )

    # Run the conda-unpack script to clean up prefixes
    # This will fix problems related to repeated "placehold" paths.
    # See https://conda.github.io/conda-pack/#commandline-usage
    unpack_script = os.path.join(pack_folder_path, "bin", "conda-unpack")
    if os.path.exists(unpack_script):
        if is_in_notebook_session() or NO_CONTAINER:
            run_command(unpack_script)
        else:
            volumes = {
                pack_folder_path: {"bind": os.path.join(DEFAULT_IMAGE_HOME_DIR, slug)}
            }
            try:
                run_container(
                    image=ML_JOB_IMAGE,
                    bind_volumes=volumes,
                    env_vars={},
                    command=os.path.join(
                        DEFAULT_IMAGE_HOME_DIR, slug, "bin/conda-unpack"
                    ),
                )
            except Exception:
                raise RuntimeError(f"Error unpacking environment {slug}.")
        if os.path.exists(os.path.join(pack_folder_path, "spark-defaults.conf")):
            if is_in_notebook_session() or NO_CONTAINER:
                if not os.path.exists(DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR):
                    os.makedirs(DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR)
                    shutil.copy(
                        os.path.join(pack_folder_path, "spark-defaults.conf"),
                        DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR,
                    )
                    for file in os.listdir(pack_folder_path):
                        if os.path.splitext(file)[1] == ".jar":
                            shutil.copy(
                                os.path.join(pack_folder_path, file),
                                DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR,
                            )
            else:
                with open(os.path.join(pack_folder_path, "spark-defaults.conf")) as f:
                    confs = f.read()
                confs = confs.replace(
                    DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR,
                    os.path.join(DEFAULT_IMAGE_CONDA_DIR, slug),
                )
                with open(
                    os.path.join(pack_folder_path, "spark-defaults.conf"), "w"
                ) as f:
                    f.write(confs)

    os.remove(pack_path)
    manifest_path = glob.glob(os.path.join(pack_folder_path, "*_manifest.yaml"))[0]
    with open(manifest_path) as f:
        env = yaml.safe_load(f.read())
    env["manifest"]["pack_uri"] = conda_uri
    with open(manifest_path, "w") as f:
        yaml.safe_dump(env, f)
    print(f"{slug} set up at {pack_folder_path}.")


def publish(**kwargs) -> None:
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    exec_config = p.config["execution"]
    if exec_config.get("environment_file", None):
        name = _get_name(exec_config.get("name"), exec_config.get("environment_file"))
        slug = _create(
            name=name,
            version=exec_config["version"],
            env_file=exec_config["environment_file"],
            conda_pack_folder=exec_config["conda_pack_folder"],
            gpu=exec_config.get("gpu", False),
            overwrite=exec_config["overwrite"],
        )
    else:
        slug = exec_config.get("slug")
    if not slug:
        raise ValueError("Please specify slug of the conda pack via `--slug`.")
    if not exec_config.get("conda_pack_os_prefix"):
        raise ValueError(
            "Please specify object storage path to save conda pack either via `ads opctl configure` or `--conda-pack-os-prefix`."
        )
    _publish(
        conda_slug=slug,
        conda_uri_prefix=exec_config["conda_pack_os_prefix"],
        conda_pack_folder=exec_config["conda_pack_folder"],
        oci_config=exec_config.get("oci_config"),
        oci_profile=exec_config.get("oci_profile"),
        overwrite=exec_config["overwrite"],
        auth_type=exec_config["auth"],
    )


def _publish(
    conda_slug: str,
    conda_uri_prefix: str,
    conda_pack_folder: str,
    oci_config: str,
    oci_profile: str,
    overwrite: bool,
    auth_type: str,
) -> None:
    """Publish a local conda pack to object storage location

    Parameters
    ----------
    conda_slug : str
        slug of conda pack
    conda_uri_prefix : str
        object storage prefix to save conda pack
    conda_pack_folder : str
        path to local conda folder
    oci_config : str
        oci config file path
    oci_profile : str
        oci config profile
    overwrite : bool
        whether to overwrite existing pack of the same slug
    auth_type : str
        authentication method

    Raises
    ------
    FileNotFoundError
        local conda pack folder not found
    FileNotFoundError
        manifest file not found inside conda pack
    RuntimeError
        IP packages found inside conda pack
    RuntimeError
        Packing conda failed
    RuntimeError
        Pack file not found
    """
    ns, bucket, prefix, _ = parse_conda_uri(conda_uri_prefix)

    # ======= check if conda pack and manifest exists ==============
    pack_folder_path = os.path.abspath(
        os.path.expanduser(os.path.join(conda_pack_folder, conda_slug))
    )
    if not os.path.exists(pack_folder_path):
        raise FileNotFoundError(
            f"Could not find environment {conda_slug} in {conda_pack_folder}."
        )
    paths = glob.glob(os.path.join(pack_folder_path, "*_manifest.yaml"))
    if len(paths) != 1:
        raise FileNotFoundError(
            "Could not locate manifest file in the provided environment."
        )
    else:
        manifest_location = paths[0]

    # ===== check if pack contains IP packages =========
    print(f"Loading environment information from {manifest_location}.")
    with open(manifest_location) as mlf:
        env = yaml.safe_load(mlf.read())
    if "IP" in env["manifest"] and env["manifest"]["IP"].lower() == "y":
        raise RuntimeError("This environment has IP restricted packages.")

    oci_auth = create_signer(auth_type, oci_config, oci_profile)
    client = OCIClientFactory(**oci_auth).object_storage
    publish_slug = conda_slug
    if not overwrite:
        existing_packs = client.list_objects(ns, bucket, prefix=prefix).data.objects
        pack_names = [os.path.basename(pack.name) for pack in existing_packs]
        while publish_slug in pack_names:
            ans = click.prompt(
                f"{conda_slug} exists in {conda_uri_prefix}. Enter a new name or overwrite (o)",
                default="o",
            )
            if ans == "o":
                break
            else:
                publish_slug = "_".join(ans.lower().split(" "))

    pack_script = os.path.join(os.path.dirname(os.path.abspath(__file__)), "pack.py")
    if is_in_notebook_session() or NO_CONTAINER:
        command = f"python {pack_script} {pack_folder_path}"
        run_command(command, shell=True)
    else:
        volumes = {
            pack_folder_path: {
                "bind": os.path.join(DEFAULT_IMAGE_HOME_DIR, conda_slug)
            },
            pack_script: {"bind": os.path.join(DEFAULT_IMAGE_HOME_DIR, "pack.py")},
        }
        command = f"python {os.path.join(DEFAULT_IMAGE_HOME_DIR, 'pack.py')} {os.path.join(DEFAULT_IMAGE_HOME_DIR, conda_slug)}"
        gpu = env["manifest"]["arch_type"] == "GPU"
        _check_job_image_exists(gpu)
        if gpu:
            image = ML_JOB_GPU_IMAGE
        else:
            image = ML_JOB_IMAGE
        try:
            run_container(
                image=image, bind_volumes=volumes, env_vars={}, command=command
            )
        except Exception:
            raise RuntimeError(f"Could not pack environment {conda_slug}.")

    pack_file = os.path.join(pack_folder_path, f"{conda_slug}.tar.gz")
    if not os.path.exists(pack_file):
        raise RuntimeError(f"Pack {pack_file} was not created.")
    pack_size = round(os.path.getsize(pack_file) / 2**20, 2)

    with open(manifest_location) as mlf:
        env = yaml.safe_load(mlf.read())
    manifest = env["manifest"]
    manifest["slug"] = conda_slug
    manifest["create_date"] = datetime.utcnow().strftime(
        "%a, %b %d, %Y, %H:%M:%S %Z UTC"
    )
    manifest["size_mb"] = pack_size
    ns, bucket, prefix, _ = parse_conda_uri(conda_uri_prefix)

    pack_uri = os.path.join(
        conda_uri_prefix,
        manifest.get("arch_type", "CPU").lower(),
        manifest["name"],
        str(manifest["version"]),
        publish_slug,
    )
    manifest["pack_path"] = os.path.join(
        prefix,
        manifest.get("arch_type", "CPU").lower(),
        manifest["name"],
        str(manifest["version"]),
        publish_slug,
    )
    manifest["pack_uri"] = pack_uri
    with open(manifest_location, "w") as f:
        yaml.safe_dump(env, f)
    if pack_size > 100:
        MultiPartUploader(
            pack_file, pack_uri, 10, oci_config, oci_profile, auth_type
        ).upload(manifest)
    else:
        with open(pack_file, "rb") as pkf:
            client.put_object(
                ns,
                bucket,
                manifest["pack_path"],
                pkf,
                opc_meta={"manifest": json.dumps(manifest)},
            )
    print(f"Conda pack {pack_uri} published.")
    os.remove(pack_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import ABC, ABCMeta

from ads.feature_engineering.feature_type.handler.feature_validator import (
    FeatureValidator,
)
from ads.feature_engineering.feature_type.handler.feature_warning import (
    FeatureWarning,
)
from ads.common.utils import camel_to_snake


class Name:
    def __get__(self, instance, owner):
        return camel_to_snake(owner.__name__)


class FeatureBaseType(type):
    """The helper metaclass to extend fucntionality of FeatureType class."""

    def __new__(cls, classname, bases, dictionary):
        dictionary["validator"] = FeatureValidator()
        dictionary["warning"] = FeatureWarning()
        return type.__new__(cls, classname, bases, dictionary)


class FeatureBaseTypeMeta(FeatureBaseType, ABCMeta):
    """The class to provide compatibility between ABC and FeatureBaseType metaclass."""


class FeatureType(ABC, metaclass=FeatureBaseTypeMeta):
    """
    Abstract case for feature types. Default class attribute include name and description.
    Name is auto generated using camel to snake conversion unless specified.
    """

    name = Name()
    description = "Base feature type."


class Tag:
    """Class for free form tags. Name must be specified."""

    def __init__(self, name: str) -> None:
        """
        Initialize a tag instance.

        Parameters
        ----------
        name: str
            The name of the tag.
        """
        self.name = name



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.preprocessing import LabelEncoder
from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)

__all__ = ["ModelEvaluator"]

DEFAULT_BIN_CLASS_METRICS = [
    "accuracy",
    "hamming_loss",
    "precision",
    "recall",
    "f1",
    "auc",
]
DEFAULT_MULTI_CLASS_METRICS = [
    "accuracy",
    "hamming_loss",
    "precision_weighted",
    "precision_micro",
    "recall_weighted",
    "recall_micro",
    "f1_weighted",
    "f1_micro",
]
DEFAULT_REG_METRICS = ["r2_score", "mse", "mae"]
DEFAULT_BIN_CLASS_LABELS_MAP = {
    "accuracy": "Accuracy",
    "hamming_loss": "Hamming distance",
    "kappa_score": "Cohen's kappa coefficient",
    "precision": "Precision",
    "recall": "Recall",
    "f1": "F1",
    "auc": "ROC AUC",
}
DEFAULT_MULTI_CLASS_LABELS_MAP = {
    "accuracy": "Accuracy",
    "hamming_loss": "Hamming distance",
    "precision_weighted": "Precision Weighted Average",
    "precision_micro": "Precision Micro Average",
    "recall_weighted": "Recall Weighted Average",
    "recall_micro": "Recall Micro Average",
    "f1_weighted": "F1 Weighted Average",
    "f1_micro": "F1 Micro Average",
}
DEFAULT_REG_LABELS_MAP = {
    "r2_score": "r-Squared Score",
    "root_mean_squared_error": "Root Mean Squared Error",
    "median_absolute_error": "Median Absolute Error",
}


class ModelEvaluator:
    """
    ModelEvaluator takes in the true and predicted values and returns a pandas dataframe

    Attributes
    ----------
    y_true : array-like object holding the true values for the model
    y_pred : array-like object holding the predicted values for the model
    model_name (str): the name of the model
    classes (list): list of target classes
    positive_class (str): label for positive outcome from model
    y_score : array-like object holding the scores for true values for the model
    metrics (dict): dictionary object holding model data

    Methods
    -------
    get_metrics()
        Gets the metrics information in a dataframe based on the number of classes
    safe_metrics_call(scoring_functions, *args)
        Applies sklearn scoring functions to parameters in args

    """

    def __init__(
        self,
        y_true,
        y_pred,
        model_name,
        classes=None,
        positive_class=None,
        y_score=None,
    ):
        self.y_true = y_true
        self.y_pred = np.squeeze(
            y_pred
        )  # This is a bug where y_pred is shape (1,n) from AutoML, rather than (n)
        self.y_score = y_score
        self.model_name = model_name
        self.classes = classes
        self.positive_class = positive_class
        self.metrics = {
            "model_name": model_name,
            "classes": classes,
            "positive_class": positive_class,
            "y_true": y_true,
            "y_pred": y_pred,
            "y_score": y_score,
        }

    def get_metrics(self):
        """
        Gets the metrics information in a dataframe based on the number of classes

        Parameters
        ----------
            self: (`ModelEvaluator` instance)
                The `ModelEvaluator` instance with the metrics.

        Returns
        -------
        :class:`pandas.DataFrame`
            Pandas dataframe containing the metrics
        """

        if self.classes is None:
            self._get_regression_metrics()
            return pd.DataFrame.from_dict(
                self.metrics, orient="index", columns=[self.model_name]
            )
        else:
            if len(self.classes) == 2:  # 'binary'
                self.positive_class = (
                    self.positive_class if self.positive_class else self.classes[0]
                )
                self._get_binary_metrics()
                return pd.DataFrame.from_dict(
                    self.metrics, orient="index", columns=[self.model_name]
                )
            else:  # 'multiclass'
                self._get_multiclass_metrics()
                return pd.DataFrame.from_dict(
                    self.metrics, orient="index", columns=[self.model_name]
                )

    def _get_general_metrics(self):
        try:
            args = [self.y_true, self.y_pred]
            kwargs = {"labels": self.classes}

            scoring_functions = {
                "classification_report": ("classification_report", len(args)),
                "kappa_score": ("cohen_kappa_score", len(args)),
                "raw_confusion_matrix": ("confusion_matrix", len(args)),
                "hinge_loss": ("hinge_loss", len(args)),
            }
            self.safe_metrics_call(scoring_functions, *args, **kwargs)

            args = [self.y_true, self.y_pred]
            scoring_functions_without_labs = {
                "accuracy": ("accuracy_score", len(args)),
                "zero_one_loss": ("zero_one_loss", len(args)),
                "hamming_loss": ("hamming_loss", len(args)),
            }
            self.safe_metrics_call(scoring_functions_without_labs, *args)

            cm = self.metrics["raw_confusion_matrix"]
            a = cm.astype("float")
            b = cm.sum(axis=1)[:, np.newaxis]
            normalized_cm = np.divide(a, b, out=np.zeros_like(a), where=b != 0).tolist()
            self.metrics["confusion_matrix"] = normalized_cm

            vc = pd.DataFrame(self.y_true).value_counts()
            self.metrics["class_balance"] = min(vc) / max(vc)
        except:
            raise ValueError(
                f"Errors arose when attempting to compute metrics. Metrics broke in the state: {self.metrics}"
            )

    def _get_binary_metrics(self):
        self._get_general_metrics()
        # add a metric to determine imbalance
        self.metrics["balanced accuracy"] = metrics.balanced_accuracy_score(
            self.y_true, self.y_pred
        )
        self.metrics["precision"] = metrics.precision_score(
            self.y_true, self.y_pred, pos_label=self.positive_class
        )
        self.metrics["recall"] = metrics.recall_score(
            self.y_true, self.y_pred, pos_label=self.positive_class
        )
        self.metrics["f1"] = metrics.f1_score(
            self.y_true, self.y_pred, pos_label=self.positive_class, average="binary"
        )

        (
            self.metrics["false_positive_rate"],
            self.metrics["true_positive_rate"],
            _,
        ) = metrics.roc_curve(self.y_true, self.y_pred, pos_label=self.positive_class)
        self.metrics["auc"] = metrics.auc(
            self.metrics["false_positive_rate"], self.metrics["true_positive_rate"]
        )

        if self.y_score is not None:
            if not all(0 >= x >= 1 for x in self.y_score):
                self.y_score = np.asarray(
                    [0 if x < 0 else 1 if x > 1 else x for x in self.y_score]
                )
            self.y_score = list(self.y_score)
            self.metrics["youden_j"] = (
                self.metrics["true_positive_rate"] - self.metrics["false_positive_rate"]
            )
            best_idx = np.argmax(self.metrics["youden_j"])
            self.metrics["roc_best_model_score"] = (
                self.metrics["false_positive_rate"][best_idx],
                self.metrics["true_positive_rate"][best_idx],
            )
            (
                self.metrics["precision_values"],
                self.metrics["recall_values"],
                _,
            ) = metrics.precision_recall_curve(
                self.y_true, self.y_score, pos_label=self.positive_class
            )
            pr_best_idx = np.argmax(
                self.metrics["precision_values"] + self.metrics["recall_values"]
            )
            self.metrics["pr_best_idx"] = pr_best_idx
            self.metrics["pr_best_model_score"] = (
                self.metrics["recall_values"][pr_best_idx],
                self.metrics["precision_values"][pr_best_idx],
            )
            self.metrics["average_precision_score"] = metrics.average_precision_score(
                self.y_true, self.y_score, pos_label=self.positive_class
            )

            self.metrics["brier score"] = metrics.brier_score_loss(
                self.y_true, self.y_score
            )
            self._get_lift_and_gain()
            # Compute KS Statistic curves
            self._binary_ks_curve()

    def _get_lift_and_gain(self):
        # make y_true a boolean vector
        y_true, y_score = np.asarray(self.y_true), np.asarray(self.y_score)
        y_true = y_true == self.positive_class

        sorted_indices = np.argsort(y_score)[::-1]
        y_true = y_true[sorted_indices]
        gains = np.cumsum(y_true)

        percentages = np.arange(start=1, stop=len(y_true) + 1)
        tp = sum(y_true)
        perfect = np.append(
            np.arange(start=1, stop=tp + 1) / float(tp), np.ones(len(y_true) - tp)
        )

        gains = gains / float(np.sum(y_true))
        percentages = percentages / float(len(y_true))

        self.metrics["cumulative_gain"] = np.insert(gains, 0, [0]) * 100
        self.metrics["percentages"] = np.insert(percentages, 0, [0]) * 100
        self.metrics["perfect_gain"] = np.insert(perfect, 0, [0]) * 100

        percentages = percentages[1:]
        self.metrics["lift"] = gains[1:] / percentages
        self.metrics["perfect_lift"] = perfect[1:] / percentages

    def _binary_ks_curve(self):
        """This function generates the points necessary to calculate the KS
        Statistic curve.
        Args:
            y_true (array-like, shape (n_samples)): True labels of the data.
            y_probas (array-like, shape (n_samples)): Probability predictions of
                the positive class.
        Returns:
            thresholds (numpy.ndarray): An array containing the X-axis values for
                plotting the KS Statistic plot.
            pct1 (numpy.ndarray): An array containing the Y-axis values for one
                curve of the KS Statistic plot.
            pct2 (numpy.ndarray): An array containing the Y-axis values for one
                curve of the KS Statistic plot.
            ks_statistic (float): The KS Statistic, or the maximum vertical
                distance between the two curves.
            max_distance_at (float): The X-axis value at which the maximum vertical
                distance between the two curves is seen.
            classes (np.ndarray, shape (2)): An array containing the labels of the
                two classes making up `y_true`.
        Raises:
            ValueError: If `y_true` is not composed of 2 classes. The KS Statistic
                is only relevant in binary classification.
        """
        y_true, y_probas = np.asarray(self.y_true), np.asarray(self.y_score)
        lb = LabelEncoder()
        encoded_labels = lb.fit_transform(y_true)
        if len(lb.classes_) != 2:
            raise ValueError(
                "Cannot calculate KS statistic for data with "
                "{} category/ies".format(len(lb.classes_))
            )
        idx = encoded_labels == 0
        data1 = np.sort(y_probas[idx])
        data2 = np.sort(y_probas[np.logical_not(idx)])

        ctr1, ctr2 = 0, 0
        thresholds, pct1, pct2 = [], [], []
        while ctr1 < len(data1) or ctr2 < len(data2):

            # Check if data1 has no more elements
            if ctr1 >= len(data1):
                current = data2[ctr2]
                while ctr2 < len(data2) and current == data2[ctr2]:
                    ctr2 += 1

            # Check if data2 has no more elements
            elif ctr2 >= len(data2):
                current = data1[ctr1]
                while ctr1 < len(data1) and current == data1[ctr1]:
                    ctr1 += 1

            else:
                if data1[ctr1] > data2[ctr2]:
                    current = data2[ctr2]
                    while ctr2 < len(data2) and current == data2[ctr2]:
                        ctr2 += 1

                elif data1[ctr1] < data2[ctr2]:
                    current = data1[ctr1]
                    while ctr1 < len(data1) and current == data1[ctr1]:
                        ctr1 += 1

                else:
                    current = data2[ctr2]
                    while ctr2 < len(data2) and current == data2[ctr2]:
                        ctr2 += 1
                    while ctr1 < len(data1) and current == data1[ctr1]:
                        ctr1 += 1

            thresholds.append(current)
            pct1.append(ctr1)
            pct2.append(ctr2)

        thresholds = np.asarray(thresholds)
        pct1 = np.asarray(pct1) / float(len(data1))
        pct2 = np.asarray(pct2) / float(len(data2))

        if thresholds[0] != 0:
            thresholds = np.insert(thresholds, 0, [0.0])
            pct1 = np.insert(pct1, 0, [0.0])
            pct2 = np.insert(pct2, 0, [0.0])
        if thresholds[-1] != 1:
            thresholds = np.append(thresholds, [1.0])
            pct1 = np.append(pct1, [1.0])
            pct2 = np.append(pct2, [1.0])

        differences = pct1 - pct2
        self.metrics["ks_statistic"], self.metrics["max_distance_at"] = (
            np.max(differences),
            thresholds[np.argmax(differences)],
        )

        (
            self.metrics["ks_thresholds"],
            self.metrics["ks_pct1"],
            self.metrics["ks_pct2"],
            self.metrics["ks_labels"],
        ) = (thresholds, pct1, pct2, lb.classes_)

    def _get_multiclass_metrics(self):
        self._get_general_metrics()
        y_true, y_pred, y_score = (
            np.asarray(self.y_true),
            np.asarray(self.y_pred),
            np.asarray(self.y_score),
        )
        if not all([x in self.classes for x in np.unique(y_true)]):
            logger.warning(
                f"There are classes in the test dataset that are not specified in the `classes` "
                f"attribute of model ({self.model_name}). This may lead to erroneous "
                f"results."
            )

        self.metrics["precision_weighted"] = metrics.precision_score(
            y_true, y_pred, labels=self.classes, average="weighted"
        )
        self.metrics["precision_micro"] = metrics.precision_score(
            y_true, y_pred, labels=self.classes, average="micro"
        )
        self.metrics["precision_by_label"] = metrics.precision_score(
            y_true, y_pred, labels=self.classes, average=None
        ).tolist()
        self.metrics["recall_weighted"] = metrics.recall_score(
            y_true, y_pred, labels=self.classes, average="weighted"
        )
        self.metrics["recall_micro"] = metrics.recall_score(
            y_true, y_pred, labels=self.classes, average="micro"
        )
        self.metrics["recall_by_label"] = metrics.recall_score(
            y_true, y_pred, labels=self.classes, average=None
        ).tolist()
        self.metrics["f1_weighted"] = metrics.f1_score(
            y_true, y_pred, labels=self.classes, average="weighted"
        )
        self.metrics["f1_micro"] = metrics.f1_score(
            y_true, y_pred, labels=self.classes, average="micro"
        )
        self.metrics["f1_by_label"] = metrics.f1_score(
            y_true, y_pred, labels=self.classes, average=None
        ).tolist()
        self.metrics["jaccard_weighted"] = metrics.jaccard_score(
            y_true, y_pred, labels=self.classes, average="weighted"
        )
        self.metrics["jaccard_micro"] = metrics.jaccard_score(
            y_true, y_pred, labels=self.classes, average="micro"
        )
        self.metrics["jaccard_by_label"] = metrics.jaccard_score(
            y_true, y_pred, labels=self.classes, average=None
        ).tolist()

        if self.y_score is not None:
            # Multiclass ROC
            (
                self.metrics["fpr_by_label"],
                self.metrics["tpr_by_label"],
                self.metrics["auc"],
                self.metrics["roc_best_model_score"],
            ) = (dict(), dict(), dict(), dict())
            for i, label in enumerate(self.classes):
                (
                    self.metrics["fpr_by_label"][i],
                    self.metrics["tpr_by_label"][i],
                    _,
                ) = metrics.roc_curve(y_true, y_score[:, i], pos_label=self.classes[i])
                self.metrics["auc"][i] = metrics.auc(
                    self.metrics["fpr_by_label"][i], self.metrics["tpr_by_label"][i]
                )
                youden_j = (
                    self.metrics["tpr_by_label"][i] - self.metrics["fpr_by_label"][i]
                )
                best_idx = np.argmax(youden_j)
                self.metrics["roc_best_model_score"][i] = (
                    self.metrics["fpr_by_label"][i][best_idx],
                    self.metrics["tpr_by_label"][i][best_idx],
                )
            # Multiclass PR
            (
                self.metrics["recall_values"],
                self.metrics["precision_values"],
                self.metrics["pr_best_model_score"],
            ) = (dict(), dict(), dict())
            for i, label in enumerate(self.classes):
                (
                    self.metrics["precision_values"][i],
                    self.metrics["recall_values"][i],
                    _,
                ) = metrics.precision_recall_curve(
                    y_true, y_score[:, i], pos_label=self.classes[i]
                )
                pr_best_idx = np.argmax(
                    self.metrics["precision_values"][i]
                    + self.metrics["recall_values"][i]
                )
                self.metrics["pr_best_model_score"][i] = (
                    self.metrics["recall_values"][i][pr_best_idx],
                    self.metrics["precision_values"][i][pr_best_idx],
                )

    @runtime_dependency(module="scipy", install_from=OptionalDependency.VIZ)
    def _get_regression_metrics(self):
        self.y_true = np.array(self.y_true)
        self.y_pred = np.array(np.squeeze(self.y_pred))
        args = [self.y_true, self.y_pred]
        scoring_functions = {
            "r2_score": ("r2_score", len(args)),
            "explained_variance_score": ("explained_variance_score", len(args)),
            "max_error": ("max_error", len(args)),
            "mae": ("mean_absolute_error", len(args)),
            "mse": ("mean_squared_error", len(args)),
            "median_absolute_error": ("median_absolute_error", len(args)),
        }

        self.safe_metrics_call(scoring_functions, *args)

        self.metrics["root_mean_squared_error"] = np.sqrt(self.metrics["mse"])
        self.metrics["residuals"] = self.y_true - self.y_pred
        self.metrics["mean_residuals"] = np.mean(self.metrics["residuals"])

        # For QQ Plot:
        portions = min(len(self.metrics["residuals"]), 100) + 1
        norm_quantiles = (np.arange(portions) / portions)[1:]
        self.metrics["norm_quantiles"] = scipy.stats.norm.ppf(norm_quantiles)
        resid_quant = [
            np.quantile(self.metrics["residuals"], p) for p in norm_quantiles
        ]
        self.metrics["residual_quantiles"] = scipy.stats.zscore(resid_quant)

    def safe_metrics_call(self, scoring_functions, *args, **kwargs):
        """Applies the sklearn function in `scoring_functions` to parameters in `args`.

        Parameters
        ----------
            scoring_functions: (dict)
                Scoring functions dictionary
            args: (keyword arguments)
                Arguments passed to the sklearn function from metrics

        Returns:
            Nothing

        Raises:
            Exception: If an error is enountered applying the sklearn function fn to arguments.
        """

        for name, (fn, n_params) in scoring_functions.items():
            try:
                if fn == "confusion_matrix":
                    self.metrics[name] = getattr(metrics, fn)(
                        **{
                            "y_true": args[0],
                            "y_pred": args[1],
                            "labels": kwargs["labels"],
                        }
                    )
                else:
                    self.metrics[name] = getattr(metrics, fn)(
                        *(args[:n_params]), **kwargs
                    )
            except Exception as e:
                self.metrics[name] = f"Error unable to compute {fn}, due to: {e}"



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import abstractmethod

from ads.common.decorator.deprecate import deprecated


class Explainer:
    """Base class for `GlobalExplainer` and `LocalExplainer`"""

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self):
        self.X_train_ = None
        self.y_train_ = None
        self.X_test_ = None
        self.y_test_ = None
        self.est_ = None
        self.mode_ = None
        self.class_names_ = None

    def setup(self, model, X_test, y_test, X_train=None, y_train=None):
        """
        Sets up required attributes for global explainer.

        Parameters
        ----------
        model : instance of `ADSModel`
            The model being explained
        X_test : pandas.DataFrame
            Test data to explain.
        y_test : pandas.Series
            Labels for test data.
        X_train : pandas.DataFrame, optional
            Training data to build explainer. Defaults to `None`.
        y_train : pandas.Series, optional
            Labels for training data. Defaults to `None`.
        """
        # assert isinstance(model, ADSModel)
        self.X_train_ = X_train if X_train is not None else X_test
        self.y_train_ = y_train if y_train is not None else y_test
        self.X_test_ = X_test
        self.y_test_ = y_test
        self.est_ = model
        if hasattr(model, "classes_") and model.classes_ is not None:
            self.mode_ = "classification"
            self.class_names_ = model.classes_
        else:
            self.mode_ = "regression"

    @property
    def X_train(self):
        """
        Training data.
        """
        return self.X_train_

    @property
    def y_train(self):
        """
        Labels for training data.
        """
        return self.y_train_

    @property
    def X_test(self):
        """
        Test data.
        """
        return self.X_test_

    @property
    def y_test(self):
        """
        Labels for test data.
        """
        return self.y_test_

    @property
    def mode(self):
        """
        Mode of explanation.
        Either `classification` or `regression`.
        """
        return self.mode_

    @property
    def est(self):
        """
        Model Estimator.
        """
        return self.est_

    @property
    def class_names(self):
        """
        Class names.
        Returns `None` for regression.
        """
        return self.class_names_


class GlobalExplainer(Explainer):
    """Abstract `GlobalExplainer` class. Must be subclassed to create instances."""

    @abstractmethod
    def compute_feature_importance(self, **kwargs):
        pass

    @abstractmethod
    def compute_partial_dependence(self, **kwargs):
        pass

    @abstractmethod
    def show_in_notebook(self):
        pass


class LocalExplainer(Explainer):
    """Abstract `LocalExplainer` class. Must be subclassed to create instances."""

    @abstractmethod
    def explain(self, **kwargs):
        pass


class WhatIfExplainer(Explainer):
    """Abstract `WhatIfExplainer` class. Must be subclassed to create instances."""

    @abstractmethod
    def explore_sample(self, **kwargs):
        pass

    @abstractmethod
    def explore_predictions(self, **kwargs):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents the ADS Feature Types Mixin class that extends
Pandas Series and Dataframe accessors.

Classes
-------
    ADSFeatureTypesMixin
        ADS Feature Types Mixin class that extends Pandas Series and Dataframe accessors.
"""
import inspect
from typing import Union

import pandas as pd
import tabulate
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.feature_type_manager import (
    FeatureTypeManager as feature_type_manager,
)
from ads.feature_engineering.feature_type_manager import _feature_type_by_dtype


class ADSFeatureTypesMixin:
    """ADS Feature Types Mixin class that extends Pandas Series and DataFrame accessors.

    Methods
    -------
    warning_registered(cls) -> pd.DataFrame
        Lists registered warnings for registered feature types.
    validator_registered(cls) -> pd.DataFrame
        Lists registered validators for registered feature types.
    help(self, prop: str = None) -> None
        Help method that prints either a table of available properties or, given a property,
        returns its docstring.
    """

    @staticmethod
    def _feature_type_by_dtype(dtype) -> FeatureType:
        """Determines feature type by DataFrame dtype.

        Parameters
        ----------
        dtype: pd.DataFrame.dtypes
            The Pandas series data type.

        Returns
        -------
        FeatureType
            The subclass of FeatureType.
        """
        return _feature_type_by_dtype(dtype)

    @staticmethod
    def _is_type_registered(feature_type: Union[FeatureType, str]) -> bool:
        """Checks if provided feature type is registered in the system.

        Parameters
        ----------
        feature_type: Union[FeatureType, str]
            The FeatureType subclass or a str indicating feature type.

        Returns
        -------
        bool
            True if provided feature type registered, False otherwise.
        """
        return feature_type_manager.is_type_registered(feature_type)

    @staticmethod
    def _get_type(feature_type: Union[FeatureType, str]) -> FeatureType:
        """Gets a feature type by class object or name.

        Parameters
        ----------
        feature_type: Union[FeatureType, str]
            The FeatureType subclass or a str indicating feature type.

        Returns
        -------
        FeatureType
            Found feature type.
        """
        return feature_type_manager.feature_type_object(feature_type)

    def warning_registered(self) -> pd.DataFrame:
        """Lists registered warnings for all registered feature types.

        Returns
        -------
        :class:`pandas.DataFrame`
            The list of registered warnings for registered feature types.

        Examples
        --------
        >>> df.ads.warning_registered()
               Column    Feature Type             Warning                    Handler
           -------------------------------------------------------------------------
           0      Age      continuous               zeros              zeros_handler
           1      Age      continuous    high_cardinality   high_cardinality_handler

        >>> df["Age"].ads.warning_registered()
               Feature Type             Warning                    Handler
           ---------------------------------------------------------------
           0     continuous               zeros              zeros_handler
           1     continuous    high_cardinality   high_cardinality_handler
        """
        common_columns = ["Feature Type", "Warning", "Handler"]
        if isinstance(self._obj, pd.DataFrame):
            result_df = pd.DataFrame((), columns=["Column"] + common_columns)
            for col in self._obj.columns:
                feature_type_df = self._obj[col].ads.warning_registered()
                feature_type_df.insert(0, "Column", col)
                result_df = result_df.append(feature_type_df)
        else:
            result_df = pd.DataFrame((), columns=common_columns)
            for feature_type in self._feature_type:
                feature_type_df = feature_type.warning.registered()
                feature_type_df.insert(0, "Feature Type", feature_type.name)
                feature_type_df = feature_type_df.rename(columns={"Name": "Warning"})
                result_df = result_df.append(feature_type_df)
        result_df.reset_index(drop=True, inplace=True)
        return result_df

    def validator_registered(self) -> pd.DataFrame:
        """Lists registered validators for registered feature types.

        Returns
        -------
        :class:`pandas.DataFrame`
            The list of registered validators for registered feature types

        Examples
        --------
        >>> df.ads.validator_registered()
                 Column     Feature Type        Validator                 Condition                    Handler
        ------------------------------------------------------------------------------------------------------
        0   PhoneNumber    phone_number   is_phone_number                        ()            default_handler
        1   PhoneNumber    phone_number   is_phone_number    {'country_code': '+7'}   specific_country_handler
        2    CreditCard    credit_card     is_credit_card                        ()            default_handler

        >>> df['PhoneNumber'].ads.validator_registered()
            Feature Type            Validator                 Condition                     Handler
        -------------------------------------------------------------------------------------------
        0   phone_number      is_phone_number                        ()             default_handler
        1   phone_number      is_phone_number    {'country_code': '+7'}    specific_country_handler
        """
        common_columns = ["Feature Type", "Validator", "Condition", "Handler"]
        if isinstance(self._obj, pd.DataFrame):
            result_df = pd.DataFrame((), columns=["Column"] + common_columns)
            for col in self._obj.columns:
                feature_type_df = self._obj[col].ads.validator_registered()
                feature_type_df.insert(0, "Column", col)
                result_df = result_df.append(feature_type_df)
        else:
            result_df = pd.DataFrame((), columns=common_columns)
            for feature_type in self._feature_type:
                feature_type_df = feature_type.validator.registered()
                feature_type_df.insert(0, "Feature Type", feature_type.name)
                feature_type_df = feature_type_df.rename(columns={"Name": "Validator"})
                result_df = result_df.append(feature_type_df)
        result_df.reset_index(drop=True, inplace=True)
        return result_df

    def help(self, prop: str = None) -> None:
        """Help method that prints either a table of available properties or, given an individual property,
        returns its docstring.

        Parameters
        ----------
        prop : str
            The Name of property.

        Returns
        -------
        None
            Nothing.
        """
        if prop:
            if hasattr(self, prop):
                print(inspect.getdoc(getattr(self, prop)))
                return
            print(f"Property {prop} not found.")
            return

        methods = set()
        attrs = set()

        def get_attr(c):
            for attr in dir(c):
                if not attr.startswith("__") and not attr.startswith("_"):
                    attr_obj = getattr(c, attr)
                    if callable(attr_obj):
                        doc = inspect.getdoc(attr_obj)
                        if doc and len(doc.split(".")) > 0:
                            methods.add((attr, doc.split(".")[0]))
                        else:
                            methods.add((attr, "method"))
                    else:
                        if hasattr(c.__class__, attr):
                            attr_obj = getattr(c.__class__, attr)
                            doc = inspect.getdoc(attr_obj)
                            attrs.add((attr, doc.split(".")[0]))

        get_attr(self)
        props = sorted(list(methods) + list(attrs))
        print(tabulate.tabulate(props, headers=("Property", "Description")))



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from dataclasses import dataclass, field
from typing import Dict

from ads.common.serializer import DataClassSerializable, SideEffect
from ads.model.runtime.model_deployment_details import ModelDeploymentDetails
from ads.model.runtime.model_provenance_details import ModelProvenanceDetails


@dataclass(repr=False)
class RuntimeInfo(DataClassSerializable):
    """RuntimeInfo class which is the data class represenation of the runtime yaml file."""

    model_artifact_version: str = ""
    model_deployment: ModelDeploymentDetails = field(
        default_factory=ModelDeploymentDetails
    )
    model_provenance: ModelProvenanceDetails = field(
        default_factory=ModelProvenanceDetails
    )

    @staticmethod
    def _validate_dict(obj_dict: Dict) -> bool:
        """Validate the runtime info.

        Parameters
        ----------
        obj_dict: (Dict)
            runtime content in dictionary format.

        Returns
        -------
        bool
            the validation result.
        """
        assert (
            "MODEL_ARTIFACT_VERSION" in obj_dict
        ), "runtime.yaml must have `MODEL_ARTIFACT_VERSION` field."
        assert (
            "MODEL_DEPLOYMENT" in obj_dict
        ), "runtime.yaml must have `MODEL_DEPLOYMENT` field."
        assert (
            "MODEL_PROVENANCE" in obj_dict
        ), "runtime.yaml must have `MODEL_PROVENANCE` field."
        return True

    @classmethod
    def from_env(cls) -> "RuntimeInfo":
        """Popolate the RuntimeInfo from environment variables.

        Returns
        -------
        RuntimeInfo
            A RuntimeInfo instance.
        """
        runtime_info = cls(model_artifact_version="3.0")
        return runtime_info

    def save(self):
        """Save the RuntimeInfo object into runtime.yaml file under the artifact directory.

        Returns
        -------
        None
            Nothing.
        """
        runtime_file_path = os.path.join(
            self.model_provenance.training_code.artifact_directory, "runtime.yaml"
        )
        self.to_yaml(
            uri=runtime_file_path, side_effect=SideEffect.CONVERT_KEYS_TO_UPPER.value
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.model.model_metadata import Framework
from ads.common import logger, utils
from ads.model.extractor.sklearn_extractor import SklearnExtractor
from ads.model.extractor.xgboost_extractor import XgboostExtractor
from ads.model.extractor.lightgbm_extractor import LightgbmExtractor
from ads.model.extractor.keras_extractor import KerasExtractor
from ads.model.extractor.automl_extractor import AutoMLExtractor
from ads.model.extractor.spark_extractor import SparkExtractor
from ads.model.extractor.pytorch_extractor import PytorchExtractor
from ads.model.extractor.tensorflow_extractor import TensorflowExtractor
from ads.model.extractor.huggingface_extractor import HuggingFaceExtractor


ORDERED_FRAMEWORKS = [
    "lightgbm",
    "xgboost",
    "sklearn",
    "keras",
    "tensorflow",
    "bert",
    "transformers",
    "torch",
    "spark",
    "automl",
]


class ModelInfoExtractorFactory:
    """Class that extract Model Taxonomy Metadata for all supported frameworks."""

    _estimator_map = {
        Framework.SCIKIT_LEARN: SklearnExtractor,
        Framework.XGBOOST: XgboostExtractor,
        Framework.LIGHT_GBM: LightgbmExtractor,
        Framework.KERAS: KerasExtractor,
        Framework.ORACLE_AUTOML: AutoMLExtractor,
        Framework.TENSORFLOW: TensorflowExtractor,
        Framework.PYTORCH: PytorchExtractor,
        Framework.SPARK: SparkExtractor,
        Framework.TRANSFORMERS: HuggingFaceExtractor,
    }

    @staticmethod
    def extract_info(model):
        """Extracts model taxonomy metadata.


        Parameters
        ----------
        model: [ADS model, sklearn, xgboost, lightgbm, keras, oracle_automl]
            The model object

        Returns
        -------
        `ModelTaxonomyMetadata`
            A dictionary with keys of Framework, FrameworkVersion, Algorithm, Hyperparameters of the model

        Examples
        --------
        >>> from ads.common.model_info_extractor_factory import ModelInfoExtractorFactory
        >>> metadata_taxonomy = ModelInfoExtractorFactory.extract_info(model)

        """
        from ads.common.model import ADSModel

        if isinstance(model, ADSModel):
            model = model.est
        model_framework = None
        model_bases = utils.get_base_modules(model)
        model_framework = ModelInfoExtractorFactory._get_estimator(
            model_bases=model_bases
        )
        if model_framework not in ModelInfoExtractorFactory._estimator_map:
            logger.warn(
                f"Auto-extraction of taxonomy is not supported for the provided model. "
                f"The supported models are {', '.join(ORDERED_FRAMEWORKS)}."
            )
            return None
        return ModelInfoExtractorFactory._estimator_map[model_framework](model).info()

    @staticmethod
    def _get_estimator(model_bases):
        mapping = {
            "lightgbm": Framework.LIGHT_GBM,
            "xgboost": Framework.XGBOOST,
            "sklearn": Framework.SCIKIT_LEARN,
            "keras": Framework.KERAS,
            "tensorflow": Framework.TENSORFLOW,
            "bert": Framework.BERT,
            "transformers": Framework.TRANSFORMERS,
            "torch": Framework.PYTORCH,
            "spark": Framework.SPARK,
            "automl": Framework.ORACLE_AUTOML,
        }
        for model_base in model_bases:
            for framework in ORDERED_FRAMEWORKS:
                if framework in model_base.__module__.split("."):
                    return mapping.get(framework)
        return None



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import sys

logger = logging.getLogger(__name__)
handler = logging.StreamHandler(sys.stdout)
logger.addHandler(handler)


def set_log_level(level):
    handler.setLevel(level)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
This exploratory data analysis (EDA) Mixin is used in the ADS accessor for the Pandas Dataframe.
The series of purpose-driven methods enable the data scientist to complete analysis on the dataframe.

From the accessor we have access to the pandas object the user is interacting with as well as
corresponding lists of feature types per column.
"""

import collections
import pandas as pd
import matplotlib.pyplot as plt
from ads.feature_engineering.accessor.mixin.correlation import (
    cat_vs_cat,
    cont_vs_cont,
    cat_vs_cont,
)
from ads.feature_engineering.accessor.mixin.utils import (
    _continuous_columns,
    _categorical_columns,
    _sienna_light_to_dark_color_palette,
)
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class EDAMixin:
    def feature_count(self) -> pd.DataFrame:
        """
        Counts the number of columns for each feature type and each primary feature.
        The column of primary is the number of primary feature types that is assigned to the column.

        Returns
        -------
        Dataframe with
          The number of columns for each feature type
          The number of columns for each primary feature

        Examples
        --------
        >>> df.ads.feature_type
        {'PassengerId': ['ordinal', 'category'],
        'Survived': ['ordinal'],
        'Pclass': ['ordinal'],
        'Name': ['category'],
        'Sex': ['category']}
        >>> df.ads.feature_count()
            Feature Type        Count       Primary
        0       category            3             2
        1        ordinal            3             3
        """

        feature_count = collections.defaultdict(lambda: [0, 0])

        for _, feature_types in self.feature_type.items():
            feature_count[feature_types[0]][1] += 1
            for ft in feature_types:
                feature_count[ft][0] += 1

        return pd.DataFrame(
            [
                (feature_type, count, primary)
                for feature_type, (count, primary) in feature_count.items()
            ],
            columns=["Feature Type", "Count", "Primary"],
        )

    def feature_stat(self) -> pd.DataFrame:
        """Summary statistics Dataframe provided.

        This returns feature stats on each column using FeatureType summary method.

        Examples
        --------
        >>> df = pd.read_csv('~/advanced-ds/tests/vor_datasets/vor_titanic.csv')
        >>> df.ads.feature_stat().head()
                 Column	   Metric	                Value
        0	PassengerId	    count	                891.000
        1	PassengerId	    mean	                446.000
        2	PassengerId	    standard deviation	    257.354
        3	PassengerId	    sample minimum  	    1.000
        4	PassengerId	    lower quartile	        223.500

        Returns
        -------
        :class:`pandas.DataFrame`
            Dataframe with 3 columns: name, metric, value

        """
        stats = []
        for col_name, col in self._obj.items():
            for _, row in col.ads.feature_stat().iterrows():
                stats.append([col_name, row["Metric"], row["Value"]])
        df_stats = pd.DataFrame(stats, columns=["Column", "Metric", "Value"])
        df_stats.value = df_stats.Value.round(3)
        return df_stats

    def feature_plot(self) -> pd.DataFrame:
        """For every column in the dataframe plot generate a list of summary plots based on the most
        relevant feature type.

        Returns
        -------
        :class:`pandas.DataFrame`
            Dataframe with 2 columns:
            1. Column - feature name
            2. Plot - plot object
        """
        plots = []
        for _, col in self._obj.items():
            try:
                plot = col.ads.feature_plot()
            except:
                plot = None
            plots.append([col.name, plot])
        return pd.DataFrame(plots, columns=["Column", "Plot"])

    def pearson(self) -> pd.DataFrame:
        """Generate a Pearson correlation data frame for all continuous variable pairs.

        Gives a warning for dropped non-numerical columns.

        Returns
        -------
        :class:`pandas.DataFrame`
        Pearson correlation data frame with the following 3 columns:
            1. Column 1 (name of the first continuous column)
            2. Column 2 (name of the second continuous column)
            3. Value (correlation value)

        Note
        ____
        Pairs will be replicated. For example for variables x and y, we'd have (x,y), (y,x) both with same correlation value. We'll also have (x,x) and (y,y) with value 1.0.
        """
        continuous_cols = _continuous_columns(self._obj.ads.feature_type)
        return cont_vs_cont(self._obj[continuous_cols])

    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def pearson_plot(self) -> plt.Axes:
        """Generate a heatmap of the Pearson correlation for all continuous variable pairs.

        Returns
        -------
        Plot object
            Pearson correlation plot object that can be updated by the customer
        """
        ax = plt.axes()
        df = (
            self.pearson()
            .pivot_table(index="Column 1", columns="Column 2", values="Value")
            .rename_axis("")
            .rename_axis("", axis="columns")
        )
        ax.set_title("Pearson's Correlation")
        return seaborn.heatmap(df, cmap=_sienna_light_to_dark_color_palette(), ax=ax)

    def cramersv(self) -> pd.DataFrame:
        """Generate a Cramer's V correlation data frame for all categorical variable pairs.

        Gives a warning for dropped non-categorical columns.

        Returns
        -------
        :class:`pandas.DataFrame`
            Cramer's V correlation data frame with the following 3 columns:
                1. Column 1 (name of the first categorical column)
                2. Column 2 (name of the second categorical column)
                3. Value (correlation value)
        Note
        ____
        Pairs will be replicated. For example for variables x and y, we would have (x,y), (y,x) both with same correlation value. We will also have (x,x) and (y,y) with value 1.0.
        """
        categorical_cols = _categorical_columns(self._obj.ads.feature_type)
        return cat_vs_cat(self._obj[categorical_cols])

    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def cramersv_plot(self) -> plt.Axes:
        """Generate a heatmap of the Cramer's V correlation for all categorical variable pairs.

        Gives a warning for dropped non-categorical columns.

        Returns
        -------
        Plot object
            Cramer's V correlation plot object that can be updated by the customer
        """
        ax = plt.axes()
        df = (
            self.cramersv()
            .pivot_table(index="Column 1", columns="Column 2", values="Value")
            .rename_axis("")
            .rename_axis("", axis="columns")
        )
        ax.set_title("Cramer's V")
        return seaborn.heatmap(df, cmap=_sienna_light_to_dark_color_palette(), ax=ax)

    def correlation_ratio(self) -> pd.DataFrame:
        """Generate a Correlation Ratio data frame for all categorical-continuous variable pairs.

        Returns
        -------
        :class:`pandas.DataFrame`
        Correlation Ratio correlation data frame with the following 3 columns:
            1. Column 1 (name of the first categorical/continuous column)
            2. Column 2 (name of the second categorical/continuous column)
            3. Value (correlation value)

        Note
        ____
        Pairs will be replicated. For example for variables x and y, we would have (x,y), (y,x) both with same correlation value. We will also have (x,x) and (y,y) with value 1.0.
        """
        categorical_cols = _categorical_columns(self._obj.ads.feature_type)
        continuous_cols = _continuous_columns(self._obj.ads.feature_type)
        return cat_vs_cont(self._obj, categorical_cols, continuous_cols)

    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def correlation_ratio_plot(self) -> plt.Axes:
        """Generate a heatmap of the Correlation Ratio correlation for all categorical-continuous variable
        pairs.

        Returns
        -------
        Plot object
            Correlation Ratio correlation plot object that can be updated by the customer
        """
        ax = plt.axes()
        df = (
            self.correlation_ratio()
            .pivot_table(index="Column 1", columns="Column 2", values="Value")
            .rename_axis("")
            .rename_axis("", axis="columns")
        )
        ax.set_title("Correlation Ratio")
        return seaborn.heatmap(df, cmap=_sienna_light_to_dark_color_palette(), ax=ax)

    def warning(self) -> pd.DataFrame:
        """Generates a data frame that lists feature specific warnings.

        Returns
        -------
        :class:`pandas.DataFrame`
            The list of feature specific warnings.

        Examples
        --------
        >>> df.ads.warning()
            Column    Feature Type         Warning               Message       Metric    Value
        --------------------------------------------------------------------------------------
        0      Age      continuous           Zeros      Age has 38 zeros        Count       38
        1      Age      continuous           Zeros   Age has 12.2% zeros   Percentage    12.2%
        """
        common_columns = ["Feature Type", "Warning", "Message", "Metric", "Value"]
        result_df = pd.DataFrame((), columns=["Column"] + common_columns)
        for col in self._obj.columns:
            warning_df = self._obj[col].ads.warning()
            if warning_df is not None:
                warning_df.insert(0, "Column", col)
                result_df = result_df.append(warning_df)
        return result_df.reset_index(drop=True)
